import warnings

from tala.model.semantic_object import OntologySpecificSemanticObject
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin

TOP = "top"
UP = "up"
HOW = "how"


class Action(OntologySpecificSemanticObject, AsSemanticExpressionMixin):
    @classmethod
    def create_from_json_api_data(cls, action_data, included):
        return cls(action_data["attributes"]["name"], action_data["attributes"]["ontology_name"])

    def __init__(self, value, ontology_name):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        self.value = value

    def is_action(self):
        return True

    @property
    def name(self):
        return self.value

    def get_value(self):
        warnings.warn("Action.get_value() is deprecated. Use Action.value instead.", DeprecationWarning, stacklevel=2)
        return self.value

    def is_top_action(self):
        return self.value == TOP

    def is_up_action(self):
        return self.value == UP

    def is_how_action(self):
        return self.value == HOW

    def __str__(self):
        return self.value

    def __hash__(self):
        return hash((self.ontology_name, self.value))

    def __eq__(self, other):
        try:
            return other.is_action() and other.value == self.value and other.ontology_name == self.ontology_name
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def as_json_api_dict(self):
        return {
            "data": {
                "type": "tala.model.action",
                "id": f"{self.ontology_name}:{self.name}",
                "attributes": {
                    "ontology_name": self.ontology_name,
                    "name": self.name
                }
            },
            "included": []
        }


class TopAction(Action):
    def __init__(self, ontology_name):
        Action.__init__(self, TOP, ontology_name)


class UpAction(Action):
    def __init__(self, ontology_name):
        Action.__init__(self, UP, ontology_name)


class HowAction(Action):
    def __init__(self, ontology_name):
        Action.__init__(self, HOW, ontology_name)



class UnicodifyException(Exception):
    pass


def unicodify(object):
    if isinstance(object, str):
        return "'%s'" % str(object)
    if isinstance(object, list):
        return _sequence_string(object)
    if isinstance(object, dict):
        return _dict_string(object)
    if isinstance(object, tuple):
        return _tuple_string(object)
    else:
        try:
            return str(object)
        except UnicodeDecodeError as exception:
            raise UnicodifyException("failed to unicodify object of class %s (%s)" % (object.__class__, exception))


def _sequence_string(sequence):
    return "[%s]" % ", ".join([unicodify(item) for item in sequence])


def _dict_string(dict_):
    formatted_key_value_pairs = [(unicodify(key), unicodify(value)) for key, value in list(dict_.items())]
    formatted_content = ["%s: %s" % pair for pair in formatted_key_value_pairs]
    return "{%s}" % ", ".join(formatted_content)


def _tuple_string(_tuple):
    return "(%s)" % ", ".join([unicodify(item) for item in _tuple])



import unittest

from unittest.mock import Mock

from tala.ddd.domain_manager import DomainManager, UnknownGoalException
from tala.ddd.ddd_manager import DDDManager
from tala.model.domain import Domain
from tala.model.goal import Goal


class DomainManagerTests(unittest.TestCase):
    def setUp(self):
        self.domain_manager = DomainManager(Mock(spec=DDDManager))

    def test_get_without_adding(self):
        self.given_mocked_domain_with_goal()
        self.when_getting_domain_of_goal_then_exception_is_raised()

    def when_getting_domain_of_goal_then_exception_is_raised(self):
        with self.assertRaises(UnknownGoalException):
            self.domain_manager.get_domain_of_goal(self.mock_goal)

    def test_add_and_get(self):
        self.given_mocked_domain_with_goal()
        self.given_domain_has_been_added()
        self.when_getting_domain_of_goal()
        self.then_got_domain()

    def given_mocked_domain_with_goal(self):
        self.mock_goal = Mock(spec=Goal)
        self.mock_goal.is_resolve_goal.return_value = False
        self.mock_domain = Mock(spec=Domain)
        self.mock_domain.goals = [self.mock_goal]

    def given_domain_has_been_added(self):
        self.domain_manager.add(self.mock_domain)

    def when_getting_domain_of_goal(self):
        self.result = self.domain_manager.get_domain_of_goal(self.mock_goal)

    def then_got_domain(self):
        self.assertEqual(self.mock_domain, self.result)

    def test_remove_added_domain(self):
        self.given_mocked_domain_with_goal()
        self.given_domain_has_been_added()
        self.given_removed_domain()
        self.when_getting_domain_of_goal_then_exception_is_raised()

    def given_removed_domain(self):
        self.domain_manager.remove(self.mock_domain)

    def test_removing_unknown_domain(self):
        self.given_mocked_domain_with_goal()
        self.when_removing_domain_then_exception_is_raised()

    def when_removing_domain_then_exception_is_raised(self):
        with self.assertRaises(UnknownGoalException):
            self.domain_manager.remove(self.mock_domain)

    def test_add_and_get_all(self):
        self.given_mocked_domain_with_goal()
        self.given_domain_has_been_added()
        self.when_getting_domains()
        self.then_domain_in_result()

    def when_getting_domains(self):
        self.result = self.domain_manager.domains

    def then_domain_in_result(self):
        self.assertIn(self.mock_domain, self.result)

    def test_remove_added_domain_leaves_no_domains(self):
        self.given_mocked_domain_with_goal()
        self.given_domain_has_been_added()
        self.given_removed_domain()
        self.when_getting_domains()
        self.then_result_is_empty()

    def then_result_is_empty(self):
        self.assertEqual(len(self.result), 0)



import structlog
import time
import threading
import random
import uuid

import pytest

from tala.utils.sse_client import SSEClient, ChunkJoiner
from tala.utils.func import configure_stdout_logging, getenv

logger = structlog.get_logger(__name__)
log_level = getenv("LOG_LEVEL", default="DEBUG")
configure_stdout_logging(log_level)


class TestSSEClient:
    def setup_method(self):
        pass

    def test_creation(self):
        self.given_args_for_client_creation(["", logger, "some_endpoint", 443])
        self.when_sse_client_created()
        self.then_client_created_with(logger, "some_endpoint", 443, "name_base")

    def given_args_for_client_creation(self, args):
        self._sse_args = args

    def when_sse_client_created(self):
        self._sse_client = SSEClient(*self._sse_args)

    def then_client_created_with(self, logger, endpoint, port, name_base):
        assert self._sse_client.logger == logger
        assert self._sse_client._endpoint == endpoint
        assert self._sse_client._port == port

    @pytest.mark.parametrize("chunk", ["This is an unproblematic test utterance.", "Frölunda ", "Fr\u00f6lunda"])
    def test_stream_single_chunk(self, chunk):
        self.given_sse_client_started(
            "test_client", logger, "wss://tala-sse-ng-g6bpb0cncyc4htg3.swedencentral-01.azurewebsites.net", 443
        )
        self.given_session_id("test-session-id")
        self.when_chunk_streamed(chunk)
        self.then_everything_is_ok()

    def given_sse_client_started(self, name, logger, endpoint, port):
        self._client = SSEClient(name, logger, endpoint, port)
        self._client.start()

    def given_session_id(self, id_):
        self._session_id = id_ + str(uuid.uuid4())

    def when_chunk_streamed(self, chunk):
        self._client.open_session(self._session_id, None)
        self._client.set_persona(self._session_id, "some-persona")
        self._client.set_voice(self._session_id, "some-voice")
        self._client.stream_chunk(self._session_id, chunk)
        self._client.flush_stream(self._session_id)
        self._client.close_session(self._session_id)

    def then_everything_is_ok(self):
        assert True

    def test_multi_sessions_single_client_no_message_confusion(self):
        self.given_mocked_stream_to_frontend()
        self.given_single_client()
        self.when_client_streaming_to_num_sessions(10)
        self.then_all_streams_correct()

    def given_mocked_stream_to_frontend(self):
        def mocked_stream_to_frontend(_self, streamer_session, message):
            session_id = streamer_session["session_id"]
            if session_id not in self._streams:
                self._streams[session_id] = []
            self._streams[session_id].append(message)

        self._streams = {}
        SSEClient._stream_to_frontend = mocked_stream_to_frontend

    def given_single_client(self):
        print("create client")
        self._client = SSEClient(
            "some-id", logger, "wss://tala-sse-ng-g6bpb0cncyc4htg3.swedencentral-01.azurewebsites.net", 443
        )
        print("client created")

    def when_client_streaming_to_num_sessions(self, num_sessions):
        for i in range(0, num_sessions):
            session_id = f"session-{i}"
            print("open session for session id", session_id)
            self._client.open_session(session_id)

        for round in range(0, 4):
            print("round", round)
            for i in random.sample(range(0, num_sessions), num_sessions):
                print("stream to", i)
                self._client.stream_chunk(f"session-{i}", f"session-{i} ")

        for i in range(0, num_sessions):
            print("close stream", i)
            client = self._client
            client.flush_stream(f"session-{i}")
            client.end_stream(f"session-{i}")
            client.close_session(f"session-{i}")
            print("done")

    def then_all_streams_correct(self):
        for session_id, streamed in self._streams.items():
            for item in streamed:
                if item["event"] == "STREAMING_CHUNK":
                    assert session_id + " " == item["data"]


class TestChunkJoiner:
    def test_single_chunk(self):
        self.given_joiner()
        self.given_chunks(["hej"])
        self.then_resulting_chunks_are(["hej"])

    def given_joiner(self):
        self._joiner = ChunkJoiner(logger)

    def given_chunks(self, chunks):
        def produce_chunks():
            for chunk in chunks:
                time.sleep(random.uniform(0.0, 0.2))
                self._joiner.add_chunk(chunk)
            self.when_end_chunks()

        self.producer_thread = threading.Thread(target=produce_chunks)
        self.producer_thread.start()

    def when_end_chunks(self):
        self._joiner.last_chunk_sent()

    def then_resulting_chunks_are(self, expected_chunks):
        self._result = list(self._joiner)
        assert len(expected_chunks) == len(self._result), f"{len(expected_chunks)} != {len(self._result)}"
        for expected, actual in zip(expected_chunks, self._result):
            assert expected == actual

    def test_two_chunks(self):
        self.given_joiner()
        self.given_chunks(["hej", " kalle"])
        self.then_resulting_chunks_are(["hej", " kalle"])

    def test_two_chunks_should_be_joined_if_no_initial_space(self):
        self.given_joiner()
        self.given_chunks(["hej", "san"])
        self.then_resulting_chunks_are(["hejsan"])

    def test_three_chunks_should_be_joined_if_no_initial_space(self):
        self.given_joiner()
        self.given_chunks(["hej", "san", "sa"])
        self.then_resulting_chunks_are(["hejsansa"])

    def test_two_chunks_should_be_joined_when_third_has_initial_space(self):
        self.given_joiner()
        self.given_chunks(["hej", "san", " sa", " kalle"])
        self.then_resulting_chunks_are(["hejsan", " sa", " kalle"])

    def test_naturalistic_gpt_output(self):
        self.given_joiner()
        self.given_chunks([
            "En", " lust", "j", "akt", " är", " en", " b", "åt", " som", " används", " för", " nö", "jes", "seg",
            "ling", ". "
        ])
        self.then_resulting_chunks_are([
            "En", " lustjakt", " är", " en", " båt", " som", " används", " för", " nöjessegling. "
        ])

    def test_ndg_system_case(self):
        self.given_joiner()
        self.given_chunks(["Har du några fler frågor? "])
        self.then_resulting_chunks_are(["Har du några fler frågor? "])

    def test_ndg_system_case_no_space(self):
        self.given_joiner()
        self.given_chunks(["Har du några fler frågor?"])
        self.then_resulting_chunks_are(["Har du några fler frågor?"])



import pytest

from tala.utils.pronunciation import Lexicon

BASIC_LEXICON = [{
    "type": "correction_entry",
    "id": "8726b03b272f90",
    "attributes": {
        "match": "utterance",
        "replacement": "spoken words"
    }
}, {
    "type": "correction_entry",
    "id": "8726b03b272f0",
    "attributes": {
        "match": "ddd_specific_entry",
        "replacement": "ddd_specific_replacement",
        "ddd_name": "some_ddd",
    }
}, {
    "type": "correction_entry",
    "id": "8726b03b272f05675",
    "attributes": {
        "match": "completely_specified_token",
        "replacement": "completely_specified_replacement",
        "ddd_name": "some_ddd",
        "voice": "some_voice",
        "locale": "some_locale",
    }
}]


class TestLexicon(object):
    def setup_method(self):
        pass

    def test_basic_succesful_lookup(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "utterance"})
        self.when_correction_called()
        self.then_result_is("spoken words")

    def given_lexicon(self, lexicon):
        self._lexicon = Lexicon(lexicon)

    def given_input_data(self, input_data):
        self._input_data = input_data

    def when_correction_called(self):
        self._pronunciation = self._lexicon.generate_pronunciation_text(self._input_data)

    def then_result_is(self, expected_pronunciation):
        assert expected_pronunciation == self._pronunciation

    def test_empty_string_lookup(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": ""})
        self.when_correction_called()
        self.then_result_is("")

    def test_succesful_lookup_ddd_given_general_correction(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "utterance", "ddd_name": "some_ddd"})
        self.when_correction_called()
        self.then_result_is("spoken words")

    def test_no_match_lookup(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "token_not_in_lexicon"})
        self.when_correction_called()
        self.then_result_is("token_not_in_lexicon")

    def test_ddd_based_lookup(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "ddd_specific_entry", "ddd_name": "some_ddd"})
        self.when_correction_called()
        self.then_result_is("ddd_specific_replacement")

    def test_ddd_based_lookup_wrong_ddd(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "ddd_specific_entry", "ddd_name": "some_other_ddd"})
        self.when_correction_called()
        self.then_result_is("ddd_specific_entry")

    def test_ddd_based_lookup_voice_no_impact_if_not_in_lexicon_entry(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "ddd_specific_entry", "ddd_name": "some_ddd", "voice": "some_voice"})
        self.when_correction_called()
        self.then_result_is("ddd_specific_replacement")

    def test_ddd_based_lookup_locale_no_impact_if_not_in_lexicon_entry(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"token": "ddd_specific_entry", "ddd_name": "some_ddd", "locale": "locale"})
        self.when_correction_called()
        self.then_result_is("ddd_specific_replacement")

    def test_ddd_based_lookup_all_parameters(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({
            "token": "completely_specified_token",
            "ddd_name": "some_ddd",
            "locale": "some_locale",
            "voice": "some_voice"
        })
        self.when_correction_called()
        self.then_result_is("completely_specified_replacement")

    @pytest.mark.parametrize("non_matching_param", ["ddd_name", "locale", "voice"])
    def test_ddd_based_lookup_all_parameters_fails_non_matching_param(self, non_matching_param):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({
            "token": "token_with_one_non_matching_param",
            "ddd_name": "some_ddd",
            "locale": "some_locale",
            "voice": "some_voice"
        })
        self.given_non_matching_param(non_matching_param)
        self.when_correction_called()
        self.then_result_is("token_with_one_non_matching_param")

    def given_non_matching_param(self, param):
        self._input_data[param] = "non_matching_value"

    def test_generate_replacement_dict(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"locale": "some_locale"})
        self.when_get_replacement_dict_called()
        self.then_result_is({"utterance": "spoken words"})

    def when_get_replacement_dict_called(self):
        self._pronunciation = self._lexicon.get_replacement_dict(self._input_data)

    def test_generate_replacement_dict_anything_goes(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"locale": "some_locale", "ddd_name": "other_name", "voice": "hmv"})
        self.when_get_replacement_dict_called()
        self.then_result_is({"utterance": "spoken words"})

    def test_generate_replacement_dict_ddd_and_locale(self):
        self.given_lexicon(BASIC_LEXICON)
        self.given_input_data({"locale": "some_locale", "ddd_name": "some_ddd", "voice": "hmv"})
        self.when_get_replacement_dict_called()
        self.then_result_is({'ddd_specific_entry': 'ddd_specific_replacement', 'utterance': 'spoken words'})



TIS_LOGGING_FULL = "full"
TIS_LOGGING_COMPACT = "compact"
TIS_LOGGING_DIFF = "diff"
TIS_LOGGING_NONE = "none"
VALID_TIS_LOGGING_FORMATS = [TIS_LOGGING_FULL, TIS_LOGGING_COMPACT, TIS_LOGGING_NONE]
VALID_TIS_UPDATE_FORMATS = [TIS_LOGGING_FULL, TIS_LOGGING_DIFF]
TIS_LOGGING_AUTO = "auto"



# coding: utf-8

import unittest

from tala.model.person_name import PersonName


class PersonNameTestCase(unittest.TestCase):
    def test_unicode(self):
        self.when_get_unicode_string(PersonName("Åsa"))
        self.then_result_is('person_name(Åsa)')

    def when_get_unicode_string(self, person_name):
        self._actual_result = str(person_name)

    def then_result_is(self, expected_result):
        self.assertEqual(expected_result, self._actual_result)

    def test_repr(self):
        self.when_get_repr(PersonName("Åsa"))
        self.then_result_is("PersonName(%r)" % "Åsa")

    def when_get_repr(self, person_name):
        self._actual_result = repr(person_name)

    def test_str_does_not_crash_with_unicode_string(self):
        self.when_get_str_string(PersonName("Åsa"))
        self.then_result_is('person_name(Åsa)')

    def when_get_str_string(self, person_name):
        self._actual_result = str(person_name)



from tala.model.semantic_object import SemanticObject


class ActionStatus(SemanticObject):
    DONE = "done"
    ABORTED = "aborted"


class Done(ActionStatus):
    def __eq__(self, other):
        return isinstance(other, Done)

    def as_semantic_expression(self):
        return self.DONE

    def __str__(self):
        return self.DONE

    def __repr__(self):
        return "Done()"

    def __hash__(self):
        return hash(self.DONE)

    def as_dict(self):
        return {"action_status": self.DONE}


class Aborted(ActionStatus):
    def __init__(self, reason):
        self._reason = reason

    @property
    def reason(self):
        return self._reason

    def __eq__(self, other):
        return isinstance(other, Aborted) and self.reason == other.reason

    def as_semantic_expression(self):
        return f"{self.ABORTED}({self.reason})"

    def __str__(self):
        return f"{self.ABORTED}({self.reason})"

    def __repr__(self):
        return f"Aborted({self.reason})"

    def __hash__(self):
        return hash((self.ABORTED, self.reason))

    def as_dict(self):
        return {"action_status": self.ABORTED, "reason": self.reason}



import unittest

from unittest.mock import Mock

from tala.ddd.extended_ddd import ExtendedDDD
from tala.ddd.parser import Parser
from tala.ddd.services.parameters.retriever import ParameterRetriever
from tala.ddd.services.service_interface import ServiceInterface
from tala.model.ddd import DDD
from tala.model.domain import Domain
from tala.model.ontology import Ontology


class TestExtendedDDD(unittest.TestCase):
    def setUp(self):
        self._extended_ddd = None
        self._mocked_ontology = None
        self._mocked_parser = None

    def test_reset_resets_ontology(self):
        self.given_mocked_ontology()
        self.given_extended_ddd_created()
        self.when_calling_reset()
        self.then_ontology_is_reset()

    def given_mocked_ontology(self, individuals=None):
        self._mocked_ontology = Mock(spec=Ontology)
        self._mocked_ontology.get_individuals.return_value = individuals or {}

    def given_extended_ddd_created(self):
        ddd = self._create_ddd()
        self._extended_ddd = self._create_extended_ddd(ddd)

    def _create_extended_ddd(self, ddd):
        return ExtendedDDD(ddd, parameter_retriever=Mock(spec=ParameterRetriever), parser=self._mocked_parser)

    def _create_ddd(self):
        return DDD(
            "a_ddd",
            ontology=self._mocked_ontology,
            domain=Mock(spec=Domain),
            service_interface=Mock(spec=ServiceInterface)
        )

    def when_calling_reset(self):
        self._extended_ddd.reset()

    def then_ontology_is_reset(self):
        self._mocked_ontology.reset.assert_called_once_with()

    def test_reset_clears_parser(self):
        self.given_mocked_ontology()
        self.given_mocked_parser()
        self.given_extended_ddd_created()
        self.when_calling_reset()
        self.then_parser_is_cleared()

    def given_mocked_parser(self):
        self._mocked_parser = Mock(spec=Parser)

    def then_parser_is_cleared(self):
        self._mocked_parser.clear.assert_called_once_with()



import warnings

from tala.utils.as_json import AsJSONMixin
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin


class Set(AsSemanticExpressionMixin, AsJSONMixin):
    def __init__(self, content=None):
        if content is None:
            content = []
        super(Set, self).__init__()
        self.content = []
        for x in content:
            self.add(x)

    def __str__(self):
        return "{" + ", ".join(map(str, self.content)) + "}"

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.content)

    def as_dict(self):
        return {
            "set": self.content,
        }

    def __eq__(self, other):
        try:
            equality = self.is_subset_of(other) and other.is_subset_of(self)
            return equality
        except AttributeError:
            return False
        except TypeError:
            return False

    def __hash__(self):
        return hash(self.__class__.__name__) + hash(self.content)

    def add(self, element):
        if element not in self.content:
            self.content.append(element)

    def remove(self, element):
        self.content.remove(element)

    def remove_if_exists(self, element):
        if element in self.content:
            self.remove(element)

    def is_subset_of(self, other):
        for item in self.content:
            if item not in other:
                return False
        return True

    def __len__(self):
        return len(self.content)

    def is_empty(self):
        return len(self) == 0

    def isEmpty(self):
        warnings.warn("Set.isEmpty() is deprecated. Use Set.is_empty() instead.", DeprecationWarning, stacklevel=2)
        return self.is_empty()

    def clear(self):
        self.content = []

    def __iter__(self):
        return self.content.__iter__()

    def union(self, other):
        union_set = Set()
        for item in self:
            union_set.add(item)
        for item in other:
            union_set.add(item)
        return union_set

    def extend(self, other):
        for item in other:
            self.add(item)

    def intersection(self, other):
        intersection_set = Set()
        for item in self:
            if item in other:
                intersection_set.add(item)
        return intersection_set



import copy
from datetime import timedelta, datetime
import json
import uuid
import time

from azure.data.tables import TableServiceClient

from tala.utils import user_rates_constants as constants

DEFAULT_AGE_LIMIT = 60


class NoConnectionToDB(Exception):
    pass


class OfferNotInDB(Exception):
    pass


class AbstractTableHandler():
    def __init__(self, connection_str: str):
        try:
            table_service = TableServiceClient.from_connection_string(conn_str=connection_str)
        except Exception:
            raise NoConnectionToDB("Table Service Client could not be reached.")
        self._table_client = table_service.get_table_client(table_name=constants.USER_RATES)

    def query_user_id(self, user_id: str):
        return self._query_entities(constants.OFFER_AUTHOR_ID, user_id)

    def query_author_user_id(self, author_user_id: str):
        return self._query_entities(constants.AUTHOR_ID, author_user_id)

    def query_offer_id(self, offer_id: str):
        return self._query_entities(constants.OFFER_ID, offer_id)

    def update_author_quota(self, author_user_id: str, author_quota: int):
        entity = self.query_author_user_id(author_user_id)[0]
        entity[constants.AUTHOR_QUOTA] = author_quota
        self._update_entity(entity)

    def _create_entity(self, entity: dict):
        table_entity = self._dump_json_fields(copy.deepcopy(entity))
        self._table_client.create_entity(table_entity)

    def _dump_json_fields(self, entity: dict):
        for field in self.fields_to_jsonify:
            try:
                entity[field] = json.dumps(entity[field])
            except KeyError:
                pass
        return entity

    def _load_json_fields(self, entity: dict):
        for field in self.fields_to_jsonify:
            try:
                entity[field] = json.loads(entity[field])
            except KeyError:
                pass
        return entity

    def _increment_calls(self, key: str, entity: dict):
        if entity:
            entity[key] += 1
            self._update_entity(entity)

    def _query_entities(self, key: str, value: str):
        filters = f"PartitionKey eq '{self.partition_key}' and {key} eq '{value}'"
        entities = list(self._table_client.query_entities(filters))
        for entity in entities:
            self._load_json_fields(entity)
        return entities

    def _update_entity(self, entity: dict):
        table_entity = self._dump_json_fields(copy.deepcopy(entity))
        self._table_client.update_entity(table_entity)


class HandlerUserRates(AbstractTableHandler):
    partition_key = constants.HANDLER_DATA
    fields_to_jsonify = [constants.CALLS_LAST_PERIOD]

    def create_entity(self, offer_id: str, user_id: str, offer_quota: int):
        entity = self._make_new_entity(offer_id, user_id, offer_quota)
        self._create_entity(entity)

    def create_entity_author_quota(self, user_id: str, author_quota: int):
        entity = self._make_new_entity_author_quota(user_id, author_quota)
        self._create_entity(entity)

    def query_offer_id_and_garbage_collect(self, offer_id: str, age_limit=DEFAULT_AGE_LIMIT):
        def garbage_collect(entity, age_limit=DEFAULT_AGE_LIMIT):
            to_remove = []
            now = datetime.now()
            for timestamp in entity.get(constants.CALLS_LAST_PERIOD, []):
                if datetime.fromtimestamp(timestamp) + timedelta(minutes=age_limit) < now:
                    to_remove.append(timestamp)
                else:
                    break
            for item in to_remove:
                entity[constants.CALLS_LAST_PERIOD].remove(item)

        entities = self.query_offer_id(offer_id)
        for entity in entities:
            garbage_collect(entity, age_limit)
            self._update_entity(entity)
        return entities

    def increment_num_calls(self, offer_id, age_limit=DEFAULT_AGE_LIMIT):
        def increment_calls(entity):
            if entity[constants.NUM_CALLS] < 0:
                pass
            else:
                entity[constants.NUM_CALLS] += 1

        def log_call_in_last_period(entity):
            try:
                entity[constants.CALLS_LAST_PERIOD].append(time.time())
            except KeyError:
                entity[constants.CALLS_LAST_PERIOD] = [time.time()]

        def get_entry_for_offer_id(offer_id):
            try:
                return self.query_offer_id(offer_id)[0]
            except IndexError:
                raise OfferNotInDB(f"Offer {offer_id} was not found in DB")
            except TypeError:
                raise OfferNotInDB(f"Offer {offer_id} was not found in DB")

        entity = get_entry_for_offer_id(offer_id)
        increment_calls(entity)
        log_call_in_last_period(entity)
        self._update_entity(entity)

    def update_offer_quota(self, offer_id: str, offer_quota: int):
        entity = self.query_offer_id(offer_id)[0]
        entity[constants.OFFER_QUOTA] = offer_quota
        self._update_entity(entity)

    def _make_new_entity(self, offer_id: str, user_id: str, offer_quota: int):
        return {
            constants.PARTITION_KEY: self.partition_key,
            constants.ROW_KEY: str(uuid.uuid4()),
            constants.NUM_CALLS: 0,
            constants.CALLS_LAST_PERIOD: [],
            constants.OFFER_ID: offer_id,
            constants.OFFER_AUTHOR_ID: user_id,
            constants.OFFER_QUOTA: offer_quota
        }

    def _make_new_entity_author_quota(self, user_id: str, author_quota: int):
        return {
            constants.PARTITION_KEY: self.partition_key,
            constants.ROW_KEY: str(uuid.uuid4()),
            constants.AUTHOR_ID: user_id,
            constants.AUTHOR_QUOTA: author_quota
        }


class BuddyGeneratorUserRates(AbstractTableHandler):
    partition_key = constants.BUDDY_GENERATOR_DATA
    fields_to_jsonify = []

    def create_entity(self, user_id: str, author_quota: int):
        entity = self._make_new_entity(user_id, author_quota)
        self._create_entity(entity)

    def increment_num_calls(self, user_id: str):
        entities = self.query_author_user_id(user_id)
        self._increment_calls(constants.NUM_CALLS, entities[0])

    def _make_new_entity(self, author_user_id: str, author_quota: int):
        return {
            constants.PARTITION_KEY: self.partition_key,
            constants.ROW_KEY: str(uuid.uuid4()),
            constants.NUM_CALLS: 0,
            constants.AUTHOR_ID: author_user_id,
            constants.AUTHOR_QUOTA: author_quota
        }



from threading import Timer

from tala.utils.observable import Observable

PASSIVE = "PASSIVE"


class PassivityTimer(Observable):
    def __init__(self):
        Observable.__init__(self)
        self._timer = None

    def start(self, duration):
        self._timer = Timer(duration, function=lambda: self.set(PASSIVE))
        self._timer.start()

    def stop(self):
        if self._timer:
            self._timer.cancel()
            self._timer = None



import pytest

from tala.ddd.json_parser import NonCheckingJSONParser
from tala.model import plan_item


class TestNonCheckingJsonParser:
    def setup_method(self):
        self._parser = NonCheckingJSONParser()

    def test_parse_log_plan_item_old_format(self):
        self._given_log_plan_item_from_old_json("test message")
        self._when_parsing_item_from_item_as_json()
        self._then_items_are_equal()

    def _given_log_plan_item_from_old_json(self, message):
        plan_item_as_json = {plan_item.TYPE_LOG: message}
        self._plan_item = self._parser.parse_plan_item(plan_item_as_json)

    def _when_parsing_item_from_item_as_json(self):
        self._parse_result = self._parser.parse_plan_item(self._plan_item.as_json())

    def _then_items_are_equal(self):
        assert self._plan_item == self._parse_result

    @pytest.mark.parametrize("log_level", plan_item.Log.LOG_LEVELS)
    def test_log_level_reflected_when_parsing_log_plan_item(self, log_level):
        self._given_plan_item_as_dict({'log': {'message': 'test message', 'level': log_level}})
        self._when_parsing_dict_as_plan_item()
        self._then_item_has_log_level(log_level)

    def _given_plan_item_as_dict(self, item):
        self._dict_to_parse = item

    def _when_parsing_dict_as_plan_item(self):
        self._parse_result = self._parser.parse_plan_item(self._dict_to_parse)

    def _then_item_has_log_level(self, level):
        assert self._parse_result.level == level

    def test_unexpected_log_level_raises_exception(self):
        self._given_plan_item_as_dict({'log': {'message': 'test message', 'level': "kalle_kula"}})
        with pytest.raises(plan_item.UnexpectedLogLevelException):
            self._when_parsing_dict_as_plan_item()



import os

from tala.model.ddd import DDD
from tala.ddd.ddd_xml_compiler import DDDXMLCompiler, DomainCompiler as DomainXmlCompiler
from tala.ddd.parser import Parser
from tala.model.domain import Domain
from tala.model.ontology import Ontology
from tala.utils import chdir


class DDDLoaderException(Exception):
    pass


class DDDLoader(object):
    def __init__(self, name, ddd_config):
        super(DDDLoader, self).__init__()
        self._name = name
        self._ddd_config = ddd_config
        self._xml_compiler = DDDXMLCompiler()

    def _compile_ontology(self):
        ontology_xml = self._load_xml_resource("ontology.xml")
        ontology_args = self._xml_compiler.compile_ontology(ontology_xml)
        ontology = Ontology(**ontology_args)
        return ontology

    def _compile_service_interface(self):
        service_interface_xml = self._load_xml_resource("service_interface.xml")
        service_interface = self._xml_compiler.compile_service_interface(service_interface_xml)
        return service_interface

    def _compile_domain(self, ontology, parser, service_interface):
        domain_args = self._domain_as_dict(ontology, parser)
        domain = Domain(ontology=ontology, **domain_args)
        return domain

    def _domain_as_dict(self, ontology, parser):
        domain_xml = self._load_xml_resource("domain.xml")
        domain_as_dict = self._xml_compiler.compile_domain(self._name, domain_xml, ontology, parser)
        return domain_as_dict

    def _load_xml_resource(self, resource_name):
        if os.path.exists(resource_name):
            with open(resource_name, "rb") as f:
                return f.read()
        else:
            raise DDDLoaderException("Expected '%s' to exist but it does not." % resource_name)

    def _find_domain_name(self):
        domain_xml = self._load_xml_resource("domain.xml")
        name = DomainXmlCompiler().get_name(domain_xml)
        return name

    def load(self):
        path = os.path.join(os.getcwd(), self._name)

        with chdir.chdir(self._name):
            ontology = self._compile_ontology()
            domain_name = self._find_domain_name()
            parser = Parser(self._name, ontology, domain_name)
            service_interface = self._compile_service_interface()
            domain = self._compile_domain(ontology, parser, service_interface)

        return DDD(self._name, ontology, domain, service_interface, path)



import json
import datetime
from pathlib import Path


class ConfigNotFoundException(Exception):
    def __init__(self, message, config_path):
        self._config_path = config_path
        super(ConfigNotFoundException, self).__init__(message)

    @property
    def config_path(self):
        return self._config_path


class BackendConfigNotFoundException(ConfigNotFoundException):
    pass


class DddConfigNotFoundException(ConfigNotFoundException):
    pass


class DeploymentsConfigNotFoundException(ConfigNotFoundException):
    pass


class UnexpectedConfigEntriesException(Exception):
    pass


class UnexpectedRASALanguageException(Exception):
    pass


class MissingMandatoryConfigEntryException(Exception):
    pass


class AbstractConfigFieldDefinition(object):
    @property
    def optional(self):
        raise NotImplementedError()


class OptionalConfigField(AbstractConfigFieldDefinition):
    def __init__(self, default_value):
        self._default_value = default_value

    @property
    def optional(self):
        return True

    @property
    def default_value(self):
        return self._default_value


class MandatoryConfigField(AbstractConfigFieldDefinition):
    @property
    def optional(self):
        return False


class Config(object):
    def __init__(self, path=None):
        path = path or self.default_name()
        self._path = Path(path)

    @property
    def _absolute_path(self):
        return Path.cwd() / self._path

    @classmethod
    def default_config(cls, **kwargs):
        def get_value(key, field_definition):
            if key in kwargs:
                return kwargs[key]
            elif field_definition.optional:
                return field_definition.default_value
            else:
                raise MissingMandatoryConfigEntryException(
                    f"Expected mandatory entry '{key}' but it is missing among entries {list(kwargs.keys())}."
                )

        return {key: get_value(key, field_definition) for key, field_definition in cls.fields().items()}

    def read(self):
        if not self._path.exists():
            self._handle_non_existing_config_file()
        self._potentially_update_and_backup_config()
        config = self._config_from_file()
        config = self._potentially_update_with_values_from_parent(config)
        self._raise_exception_if_mandatory_field_is_missing(config)
        self._set_default_values_for_missing_optional_entries(config)
        return config

    def _handle_non_existing_config_file(self):
        self._raise_config_not_found_exception()

    def _write(self, config):
        self._write_to_file(config, self._path)

    @classmethod
    def write_default_config(cls, path=None, **kwargs):
        path = path or cls.default_name()
        cls._write_to_file(cls.default_config(**kwargs), Path(path))

    def _write_backup(self, config):
        path = Path(self.back_up_name())
        self._write_to_file(config, path)

    @staticmethod
    def _write_to_file(config, path):
        with path.open(mode="w") as f:
            string = json.dumps(config, sort_keys=True, indent=4, separators=(',', ': '), ensure_ascii=False)
            f.write(str(string))

    def _config_from_file(self):
        with self._path.open(mode="r") as f:
            return json.load(f)

    def _raise_exception_if_mandatory_field_is_missing(self, config):
        for key, field_definition in self.fields().items():
            if not field_definition.optional and key not in config:
                raise MissingMandatoryConfigEntryException(
                    f"Expected mandatory entry '{key}' in '{self._path}' but it is missing among entries "
                    f"{list(config.keys())}."
                )

    def _potentially_update_and_backup_config(self):
        config = self._config_from_file()
        config = self._potentially_update_with_values_from_parent(config)
        expected_entries = set(self.fields().keys())
        expected_entries.add("overrides")
        unexpected_entries = sorted(set(config.keys()).difference(expected_entries))
        if unexpected_entries:
            self._write_backup(config)
            self._write(config)
            raise UnexpectedConfigEntriesException(
                f"Parameters {unexpected_entries} are unexpected in '{self._absolute_path}'. "
                f"The config was updated and the previous config was backed up in '{self.back_up_name()}'."
            )

    def _potentially_update_with_values_from_parent(self, config):
        parent_path = config.get("overrides")
        if parent_path:
            config = self._update_with_values_from_parent(config, parent_path)
        return config

    def _update_with_values_from_parent(self, config, parent_path):
        parent_config_object = self.__class__(parent_path)
        parent_config = parent_config_object.read()
        parent_config.update(config)
        return parent_config

    def _set_default_values_for_missing_optional_entries(self, config):
        for key, field_definition in self.fields().items():
            if field_definition.optional and key not in config:
                config[key] = field_definition.default_value

    def back_up_name(self):
        return "{}.backup".format(self._path)

    @staticmethod
    def default_name():
        raise NotImplementedError()

    @staticmethod
    def fields():
        raise NotImplementedError()

    def _raise_config_not_found_exception(self):
        raise NotImplementedError()


class BackendConfig(Config):
    DEFAULT_RERANK_AMOUNT = 0.2
    DEFAULT_INACTIVE_SECONDS_ALLOWED = datetime.timedelta(hours=2).seconds
    DEFAULT_RESPONSE_TIMEOUT = 2.5
    DEFAULT_CONFIDENCE_THRESHOLDS = {"TRUST": 0.3, "ACKNOWLEDGE": 0.15, "CHECK": 0.1}
    DEFAULT_PREDICTION_CONFIDENCE_THRESHOLDS = {"TRUST": 1.0, "ACKNOWLEDGE": 0.8, "CHECK": 0.6}
    DEFAULT_SHORT_TIMEOUT = 1.0
    DEFAULT_MEDIUM_TIMEOUT = 2.0
    DEFAULT_LONG_TIMEOUT = 5.0

    @staticmethod
    def default_name():
        return "backend.config.json"

    @staticmethod
    def fields():
        return {
            "use_recognition_profile":
            OptionalConfigField(default_value=False),
            "ddds":
            MandatoryConfigField(),
            "active_ddd":
            OptionalConfigField(default_value="none"),
            "supported_languages":
            OptionalConfigField(default_value="none"),
            "asr":
            OptionalConfigField(default_value="none"),
            "repeat_questions":
            OptionalConfigField(default_value=True),
            "rerank_amount":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_RERANK_AMOUNT),
            "inactive_seconds_allowed":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_INACTIVE_SECONDS_ALLOWED),
            "response_timeout":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_RESPONSE_TIMEOUT),
            "confidence_thresholds":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_CONFIDENCE_THRESHOLDS),
            "confidence_prediction_thresholds":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_PREDICTION_CONFIDENCE_THRESHOLDS),
            "short_timeout":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_SHORT_TIMEOUT),
            "medium_timeout":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_MEDIUM_TIMEOUT),
            "long_timeout":
            OptionalConfigField(default_value=BackendConfig.DEFAULT_LONG_TIMEOUT),
        }

    @classmethod
    def write_default_config(cls, ddd_name, path=None):
        path = path or cls.default_name()
        ddd_name = ddd_name or ""
        cls._write_to_file(cls.default_config(ddds=[ddd_name], active_ddd=ddd_name), Path(path))

    def _raise_config_not_found_exception(self):
        raise BackendConfigNotFoundException(
            "Expected backend config '{}' to exist but it was not found.".format(self._absolute_path),
            self._absolute_path
        )


class DddConfig(Config):
    @staticmethod
    def default_name():
        return "ddd.config.json"

    @staticmethod
    def fields():
        return {
            "device_module": OptionalConfigField(default_value=None),
            "rasa_nlu": OptionalConfigField(default_value={}),
            "use_rgl": OptionalConfigField(default_value=False),
            "use_third_party_parser": OptionalConfigField(default_value=False),
            "word_list": OptionalConfigField(default_value="word_list.txt"),
        }

    def _raise_config_not_found_exception(self):
        raise DddConfigNotFoundException(
            "Expected DDD config '{}' to exist but it was not found.".format(self._absolute_path), self._absolute_path
        )

    def read(self):
        config = super(DddConfig, self).read()
        return config


class DeploymentsConfig(Config):
    @staticmethod
    def default_name():
        return "deployments.config.json"

    @staticmethod
    def fields():
        return {"dev": OptionalConfigField(default_value="https://127.0.0.1:9090/interact")}

    def _raise_config_not_found_exception(self):
        message = "Expected deployments config '{}' to exist but it was not found.".format(self._absolute_path)
        raise DeploymentsConfigNotFoundException(message, self._absolute_path)

    def _potentially_update_and_backup_config(self):
        pass

    def get_url(self, candidate_deployment):
        if not self._path.exists():
            return candidate_deployment
        config = self.read()
        if candidate_deployment in config:
            return config[candidate_deployment]
        return candidate_deployment


class OverriddenDddConfig(object):
    def __init__(self, ddd_name, path):
        self._ddd_name = ddd_name
        self._path = path

    @property
    def ddd_name(self):
        return self._ddd_name

    @property
    def path(self):
        return self._path



import copy

from tala.model.action import Action
from tala.testing.lib_test_case import LibTestCase


class ActionTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_create(self):
        self.assertTrue(self.buy_action.is_action())
        self.assertEqual("buy", self.buy_action.value)

    def test_unicode(self):
        self.assertEqual("buy", str(self.buy_action))

    def test_equality(self):
        action = self.buy_action
        identical_action = copy.copy(self.buy_action)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(action, identical_action)

    def test_inequal_when_only_ontology_name_differs(self):
        action_name = "action_name"
        action = Action(action_name, "an_ontology")
        other = Action(action_name, "other_ontology")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(action, other)

    def test_inequality(self):
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.buy_action, self.top_action)

    def test_is_action(self):
        action = self.buy_action
        non_action = self.proposition_dest_city_paris
        self.assertTrue(action.is_action())
        self.assertFalse(non_action.is_action())

    def test_actions_are_hashable(self):
        {self.buy_action}



import unittest

from tala.ddd.services.parameters.formatting import ParameterBindingsFormatter
from tala.ddd.services.parameters.binding import SingleInstanceParameterBinding
from tala.ddd.services.service_interface import ServiceParameter


class FormattingTestCase(unittest.TestCase):
    def test_to_proposition_list_base_case(self):
        self.given_bindings([
            SingleInstanceParameterBinding(ServiceParameter("mock_predicate_name"), "mock_value", "mock_proposition")
        ])
        self.when_to_proposition_list_is_called()
        self.then_result_is(["mock_proposition"])

    def given_bindings(self, bindings):
        self._bindings = bindings

    def then_result_is(self, expected_result):
        self.assertEqual(expected_result, self._actual_result)

    def when_to_proposition_list_is_called(self):
        self._actual_result = ParameterBindingsFormatter.to_proposition_list(self._bindings)

    def test_to_proposition_list_excludes_optional_unbound_propositions(self):
        self.given_bindings([
            SingleInstanceParameterBinding(
                ServiceParameter("mock_predicate_name", is_optional=True), value=None, proposition=None
            )
        ])
        self.when_to_proposition_list_is_called()
        self.then_result_is([])

    def test_to_values_base_case(self):
        self.given_bindings([
            SingleInstanceParameterBinding(ServiceParameter("mock_predicate_name"), "mock_value", "mock_proposition")
        ])
        self.when_to_values_is_called()
        self.then_result_is(["mock_value"])

    def when_to_values_is_called(self):
        self._actual_result = ParameterBindingsFormatter.to_values(self._bindings)

    def test_result_has_values_includes_optional_unknown_facts(self):
        self.given_bindings([
            SingleInstanceParameterBinding(
                ServiceParameter("dest_city", is_optional=True), value=None, proposition=None
            )
        ])
        self.when_to_values_is_called()
        self.then_result_is([None])



class CacheMethod:
    def __init__(self, instance, method):
        self._method = method
        self._cache = {}

        def cached_method(*args):
            if args in self._cache:
                return self._cache[args]
            else:
                value = self._method.__call__(*args)
                self._cache[args] = value
                return value

        setattr(instance, method.__name__, cached_method)

    def __str__(self):
        return "CacheMethod(%s, _cache=%s)" % (self._method, self._cache)

    def clear(self):
        self._cache = {}



from tala.model.interpretation import Interpretation


class TestInterpretation:
    def test_create_ddd_specific_user_move_from_json(self):
        self.given_interpretation_as_dict({
            'moves': [{
                'ddd': 'some-ddd',
                'perception_confidence': 1.0,
                'semantic_expression': 'ask(?X.qna_response(X))',
                'understanding_confidence': 1.0,
            }],
            'utterance': 'some-utterance',
            'perception_confidence': 1.01,
            'modality': 'speech'
        })
        self.when_create_interpretation()
        self.then_interpretation_as_dict_equals_original_dict()

    def given_interpretation_as_dict(self, json_dict):
        self._interpretation_as_json_dict = json_dict

    def when_create_interpretation(self):
        self._interpretation = Interpretation.from_dict(self._interpretation_as_json_dict)

    def then_interpretation_as_dict_equals_original_dict(self):
        assert self._interpretation_as_json_dict == self._interpretation.as_dict()



from collections import namedtuple


class SingleInstanceParameterBinding(
    namedtuple("SingleInstanceParameterBinding", ["parameter", "value", "proposition"])
):
    @property
    def propositions(self):
        return [self.proposition]

    @property
    def is_multiple_instance_binding(self):
        return False


class MultiInstanceParameterBinding(
    namedtuple("MultiInstanceParameterBinding", ["parameter", "values", "propositions"])
):
    @property
    def value(self):
        return self.values

    @property
    def is_multiple_instance_binding(self):
        return True



from tala.testing.interaction.comparison import StringComparison


class TestStringComparison:
    def test_basic_comparison_success(self):
        self.given_actual_string("hej")
        self.given_expected_string("hej")
        self.when_comparing()
        self.then_comparison_is_successful()

    def given_actual_string(self, string):
        self._actual_string = string

    def given_expected_string(self, string):
        self._expected_string = string

    def when_comparing(self):
        self._comparison = StringComparison(self._actual_string, self._expected_string).match()

    def then_comparison_is_successful(self):
        assert self._comparison is not None
        assert self._comparison.group(0) == self._actual_string

    def test_basic_comparison_success_wildcard(self):
        self.given_actual_string("hej")
        self.given_expected_string("h*j")
        self.when_comparing()
        self.then_comparison_is_successful()

    def test_basic_comparison_failure(self):
        self.given_actual_string("haj")
        self.given_expected_string("hej")
        self.when_comparing()
        self.then_comparison_is_failing()

    def then_comparison_is_failing(self):
        assert self._comparison is None

    def test_basic_comparison_failure_wildcard(self):
        self.given_actual_string("haj")
        self.given_expected_string("he*")
        self.when_comparing()
        self.then_comparison_is_failing()

    def test_succesful_multi_alt_comparison(self):
        self.given_actual_string("hej")
        self.given_expected_string(["haj", "hej"])
        self.when_comparing()
        self.then_comparison_is_successful()

    def test_succesful_multi_alt_comparison_wildcard(self):
        self.given_actual_string("hej")
        self.given_expected_string(["*aj", "*ej"])
        self.when_comparing()
        self.then_comparison_is_successful()

    def test_failing_multi_alt_comparison(self):
        self.given_actual_string("huj")
        self.given_expected_string(["haj", "hej"])
        self.when_comparing()
        self.then_comparison_is_failing()

    def test_failing_multi_alt_comparison_wildcard(self):
        self.given_actual_string("huj")
        self.given_expected_string(["ha*", "he*"])
        self.when_comparing()
        self.then_comparison_is_failing()

    def test_basic_mismatch_output(self):
        self.given_actual_string("hej")
        self.given_expected_string("haj")
        self.when_getting_mismatch_description()
        self.then_mismatch_description_is('\nexpected: "haj"\n            ^\nbut got:  "hej"\n            ^')

    def when_getting_mismatch_description(self):

        comparison = StringComparison(self._actual_string, self._expected_string)
        comparison.match()
        self._mismatch = comparison.mismatch_description()

    def then_mismatch_description_is(self, expected):
        assert expected == self._mismatch

    def test_wildcard_mismatch_output(self):
        self.given_actual_string("hej")
        self.given_expected_string("ha*")
        self.when_getting_mismatch_description()
        self.then_mismatch_description_is('\nexpected: "ha*"\n            ^\nbut got:  "hej"\n            ^')

    def test_multi_alt_mismatch_output(self):
        self.given_actual_string("huj")
        self.given_expected_string(["haj", "hej"])
        self.when_getting_mismatch_description()
        self.then_mismatch_description_is('\nexpected: "hej"\n            ^\nbut got:  "huj"\n            ^')

    def test_multi_alt_wildcard_mismatch_output(self):
        self.given_actual_string("huj")
        self.given_expected_string(["ha*", "he*"])
        self.when_getting_mismatch_description()
        self.then_mismatch_description_is('\nexpected: "he*"\n            ^\nbut got:  "huj"\n            ^')



from tala.ddd.json_parser import JSONDDDParser
from tala.ddd.parser import Parser
from tala.ddd.services.parameters.retriever import ParameterRetriever
from tala.ddd.extended_ddd import ExtendedDDD
from tala.ddd.domain_manager import DomainManager
from tala.model.semantic_logic import SemanticLogic
from tala.model.ddd import DDD


class DDDAlreadyExistsException(Exception):
    pass


class UnexpectedDomainException(Exception):
    pass


class UnexpectedOntologyException(Exception):
    pass


class UnexpectedDDDException(Exception):
    pass


class SemanticObjectException(Exception):
    pass


def get_ddd_json_version(ddd_as_json):
    if "data" in ddd_as_json and "version:id" in ddd_as_json["data"]:
        return ddd_as_json["data"]["version:id"]
    return "1"


class DDDManager(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.ddd_names = []
        self.ddds_as_json = []
        self._ddds = {}
        self.ontologies = {}
        self.domains = {}
        self.domain_manager = DomainManager(self)
        self._ddds_of_domains = {}
        self._ddds_of_ontologies = {}
        self._semantic_logic = SemanticLogic(self)

    @property
    def semantic_logic(self):
        return self._semantic_logic

    def add(self, ddd):
        if ddd.name in self._ddds:
            raise DDDAlreadyExistsException("DDD '%s' already registered" % ddd.name)
        self.add_ontology(ddd.ontology)
        self._ddds_of_ontologies[ddd.ontology] = ddd
        self.add_domain(ddd.domain)
        self._ddds_of_domains[ddd.domain] = ddd
        self._ddds[ddd.name] = ddd

    def ensure_ddd_added(self, ddd):
        if ddd.name not in self._ddds:
            self.add(ddd)

    def add_domain(self, domain):
        self.domains[domain.get_name()] = domain
        self.domain_manager.add(domain)

    def add_ontology(self, ontology):
        self.ontologies[ontology.get_name()] = ontology

    def get_all_ddds(self):
        return list(self._ddds.values())

    def get_ddd(self, name):
        if name not in self._ddds and self.ddds_as_json:
            self._load_ddd(name)
        return self._ddds[name]

    def _load_ddd(self, name):
        if name not in self.ddd_names:
            raise UnexpectedDDDException(f"Expected one of the known DDDs {self.ddd_names}, but got '{name}'")
        if not self._is_loaded(name):
            ddd_as_json = self._get_ddd_as_json(name)
            self._parse_and_add(ddd_as_json)

    def load_ddd_for_ontology_name(self, name):
        for ddd_as_json in self.ddds_as_json:
            if get_ddd_json_version(ddd_as_json) == "2":
                ontology_name = ddd_as_json["data"]["relationships"]["ontology"]["data"]["id"]
            elif get_ddd_json_version(ddd_as_json) == "1":
                ontology_name = ddd_as_json["ontology"]["_name"]
            if ontology_name == name:
                self._parse_and_add(ddd_as_json)
                return
        raise UnexpectedDDDException(f"Expected ontology name of a known DDD ({self.ddd_names}), but got '{name}'")

    def _is_loaded(self, ddd_name):
        return ddd_name in self._ddds

    def add_ddds_as_json(self, ddd_names, ddds_as_json):
        self.ddd_names = ddd_names
        self.ddds_as_json = ddds_as_json

    def _get_ddd_as_json(self, name):
        for ddd_as_json in self.ddds_as_json:
            if get_ddd_json_version(ddd_as_json) == "2":
                if name == ddd_as_json["data"]["attributes"]["name"]:
                    return ddd_as_json
            if get_ddd_json_version(ddd_as_json) == "1":
                if name == ddd_as_json["ddd_name"]:
                    return ddd_as_json

    def _parse_and_add(self, ddd_as_json):
        if "data" in ddd_as_json and "version:id" in ddd_as_json["data"] and ddd_as_json["data"]["version:id"] == "2":
            ddd = DDD.create_from_json_api_data(ddd_as_json)
        else:
            ddd = JSONDDDParser().parse(ddd_as_json)
        parameter_retriever = ParameterRetriever(ddd.service_interface, ddd.ontology)
        parser = Parser(ddd.name, ddd.ontology, ddd.domain.name)
        extended_ddd = ExtendedDDD(ddd, parameter_retriever, parser)
        self.add(extended_ddd)

    def get_domain(self, name):
        return self.domains[name]

    def get_ontology(self, name):
        if name not in self.ontologies:
            self.load_ddd_for_ontology_name(name)
        return self.ontologies[name]

    def get_ontology_of(self, semantic_object):
        if semantic_object.is_ontology_specific():
            return self.get_ontology(semantic_object.ontology_name)
        raise SemanticObjectException(
            "This object is not ontology specific, and has no ontology information", semantic_object=semantic_object
        )

    def get_ddd_of_semantic_object(self, semantic_object):
        ontology = self.get_ontology_of(semantic_object)
        return self.get_ddd_of_ontology(ontology)

    def get_ddd_of_ontology(self, ontology):
        if ontology not in self._ddds_of_ontologies:
            raise UnexpectedOntologyException(
                "Expected to find '%s' among known ontologies %s but did not." %
                (ontology, list(self._ddds_of_ontologies.keys()))
            )
        return self._ddds_of_ontologies[ontology]

    def get_ddd_for_ontology_name(self, ontology_name):
        if ontology_name not in self.ontologies:
            self.load_ddd_for_ontology_name(ontology_name)
        ontology = self.get_ontology(ontology_name)
        return self.get_ddd_of_ontology(ontology)

    def reset_ddd(self, name):
        self._ddds[name].reset()



import unittest

from tala.model.set import Set
from tala.testing.utils import EqualityAssertionTestCaseMixin


class MockElement:
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return "MockElement(%r)" % self._name


class SetTests(unittest.TestCase, EqualityAssertionTestCaseMixin):
    def test_set(self):
        testset = Set()
        testset.add("first")
        testset.add("second")
        self.assertEqual(len(testset), 2)

    def test_ininitialize_with_content(self):
        testset = Set(["first", "second"])
        self.assertEqual(len(testset), 2)

    def test_set_eq(self):
        testset1 = Set()
        testset1.add("first")
        testset1.add("second")
        testset2 = Set()
        testset2.add("second")
        testset2.add("first")
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(testset1, testset2)

    def test_set_non_equality(self):
        testset1 = Set()
        testset1.add("third")
        testset1.add("second")
        testset2 = Set()
        testset2.add("second")
        testset2.add("first")
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(testset1, testset2)

    def test_empty_set_is_subset_of_any_set(self):
        testset = Set()
        testset.add("first")
        self.assertTrue(Set().is_subset_of(testset))

    def test_empty_set_is_subset_of_empty_set(self):
        testset = Set()
        self.assertTrue(Set().is_subset_of(testset))

    def test_single_elem_set_not_subset_of_empty_set(self):
        testset = Set()
        testset.add("first")
        self.assertFalse(testset.is_subset_of(Set()))

    def test_set_is_subset_of_itself(self):
        testset = Set()
        testset.add("first")
        self.assertTrue(testset.is_subset_of(testset))

    def test_subset_false_for_sets_with_empty_intersection(self):
        first_set = Set()
        first_set.add("first")
        second_set = Set()
        second_set.add("second")
        self.assertFalse(first_set.is_subset_of(second_set))

    def test_is_subset_of_true_for_sets_with_non_empty_intersection(self):
        first_set = Set()
        first_set.add("first")
        second_set = Set()
        second_set.add("first")
        second_set.add("second")
        self.assertTrue(first_set.is_subset_of(second_set))

    def test_empty_set_not_equals_none(self):
        empty_set = Set()
        self.assertNotEqual(None, empty_set)

    def test_empty_set_is_empty(self):
        testset = Set()
        self.assertEqual(0, len(testset))
        self.assertTrue(testset.is_empty())

    def test_non_empty_set_not_is_empty(self):
        testset = Set()
        testset.add("randomString")
        self.assertEqual(1, len(testset))
        self.assertFalse(testset.is_empty())

    def test_set_member_is_member(self):
        testset = Set()
        element = "kalle"
        testset.add(element)
        self.assertTrue(element in testset)

    def test_set_non_member_is_not_member(self):
        testset = Set()
        element = "kalle"
        self.assertFalse(element in testset)

    def test_set_iteration(self):
        testset = Set()
        testset.add("first")
        testset.add("second")
        comparisonset = set()
        for element in testset:
            comparisonset.add(element)
        self.assertTrue("first" in comparisonset)
        self.assertTrue("second" in comparisonset)

    def test_set_unicode(self):
        testset = Set()
        testset.add(MockElement("element"))
        self.assertEqual("{MockElement('element')}", str(testset))

        testset = Set()
        testset.add(MockElement("element1"))
        testset.add(MockElement("element2"))
        self.assertTrue(
            "{MockElement('element1'), MockElement('element2')}" == str(testset)
            or "{MockElement('element2'), MockElement('element1')}" == str(testset)
        )

    def test_set_property(self):
        set = Set()
        set.add("first")
        set.add("first")
        self.assertEqual(1, len(set))

    def test_remove_existing_element_from_single_element_set(self):
        set = Set(["first"])
        set.remove("first")
        self.assertEqual(0, len(set))

    def test_remove_existing_element(self):
        set = Set(["first", "second"])
        set.remove("first")
        expected_result = Set(["second"])
        self.assertEqual(expected_result, set)

    def test_remove_non_existing_element_yields_exception(self):
        set = Set(["first", "second"])
        with self.assertRaises(ValueError):
            set.remove("third")

    def test_remove_if_exists_for_existing_element(self):
        set = Set(["first", "second"])
        set.remove_if_exists("first")
        expected_result = Set(["second"])
        self.assertEqual(expected_result, set)

    def test_remove_if_exists_for_non_existing_element_has_no_effect(self):
        set = Set(["first", "second"])
        set.remove_if_exists("third")
        expected_result = Set(["first", "second"])
        self.assertEqual(expected_result, set)

    def test_union_with_empty_set(self):
        test_set = Set(["first", "second"])
        union = test_set.union(Set())
        self.assertEqual(test_set, union)

    def test_union_with_self(self):
        test_set = Set(["first", "second"])
        union = test_set.union(test_set)
        self.assertEqual(test_set, union)

    def test_union_with_empty_set_reflexive_1(self):
        test_set = Set(["first", "second"])
        union = Set().union(test_set)
        self.assertEqual(test_set, union)

    def test_union_with_empty_set_reflexive_2(self):
        first_list = ["first", "second"]
        second_list = ["third", "fourth"]
        first_set = Set(first_list)
        second_set = Set(second_list)
        union = first_set.union(second_set)
        first_list.extend(second_list)
        for item in first_list:
            self.assertTrue(item in union)

    def test_union_is_new_instance(self):
        first_set = Set()
        second_set = Set()
        union_set = first_set.union(second_set)
        first_set.add("dummy_item")
        self.assertEqual(Set(), union_set)

    def test_intersection_with_empty_set(self):
        first_set = Set()
        second_set = Set()
        intersection_set = first_set.intersection(second_set)
        self.assertEqual(first_set, intersection_set)

    def test_intersection_with_self(self):
        test_set = Set()
        intersection_set = test_set.intersection(test_set)
        self.assertEqual(test_set, intersection_set)

    def test_intersection_with_common_elements(self):
        first_set = Set([1, 2])
        second_set = Set([2, 3])
        intersection_set = first_set.intersection(second_set)

        expected_intersection = Set([2])
        self.assertEqual(expected_intersection, intersection_set)

    def test_intersection_disjoint_sets(self):
        first_set = Set([1])
        second_set = Set([2])
        intersection_set = first_set.intersection(second_set)
        expected_intersection_set = Set()
        self.assertEqual(intersection_set, expected_intersection_set)

    def test_set_in_set(self):
        set = Set()
        set_in_set = Set()
        set.add(set_in_set)

    def test_set_clear_then_add(self):
        set = Set()
        set.clear()
        set.add("first")

    def test_extend(self):
        test_set = Set(["first", "second"])
        extension = Set(["second", "third"])
        test_set.extend(extension)
        expected_result = Set(["first", "second", "third"])
        self.assertEqual(expected_result, test_set)



import copy
import datetime
import json

import pytest

from tala.utils import user_rates_constants as constants
from tala.utils.user_rates import BuddyGeneratorUserRates, HandlerUserRates, OfferNotInDB

OFFER_USAGE_QUOTA = 500
AUTHOR_USAGE_QUOTA = 2500
AUTHOR_GENERATION_QUOTA = 30


class MockHandlerUserRates(HandlerUserRates):
    def __init__(self, table_client):
        self._table_client = table_client

    def _query_entities(self, key, value):
        entities = copy.deepcopy(self._table_client.query_entities(key, value))
        for entity in entities:
            self._load_json_fields(entity)
        return entities


class MockBuddyGeneratorUserRates(BuddyGeneratorUserRates):
    def __init__(self, table_client):
        self._table_client = table_client

    def _query_entities(self, key, value):
        entities = copy.deepcopy(self._table_client.query_entities(key, value))
        for entity in entities:
            self._load_json_fields(entity)
        return entities


class MockTableClient:
    def __init__(self, rate_entries):
        self._entries = rate_entries

    def query_entities(self, key, value):
        entities = []
        for entity in self._entries:
            if key in entity and entity[key] == value:
                entities.append(entity)
        return entities

    def create_entity(self, entity):
        self._entries.append(entity)

    def update_entity(self, entity):
        for original_entity in self._entries:
            if entity["RowKey"] == original_entity["RowKey"]:
                for key in entity:
                    original_entity[key] = entity[key]
                break


def create_rate_entry(handler_calls, offer_id, row_key):
    return {
        constants.PARTITION_KEY: constants.HANDLER_DATA,
        constants.ROW_KEY: row_key,
        constants.NUM_CALLS: handler_calls,
        constants.OFFER_ID: str(offer_id),
        constants.CALLS_LAST_PERIOD: json.dumps([]),
        constants.OFFER_AUTHOR_ID: '2',
        constants.OFFER_QUOTA: OFFER_USAGE_QUOTA
    }


def create_author_quota_entry(user_id, row_key):
    return {
        constants.PARTITION_KEY: constants.HANDLER_DATA,
        constants.ROW_KEY: row_key,
        constants.AUTHOR_ID: user_id,
        constants.AUTHOR_QUOTA: AUTHOR_USAGE_QUOTA
    }


class TestHandlerUserRates:
    def test_first_call(self):
        self.given_user_rates([(0, "offer-id-2", "id-2")])
        self.given_handler()
        self.when_increment_calls("offer-id-2")
        self.then_num_calls_are("offer-id-2", 1)

    def test_total_calls_incremented(self):
        self.given_user_rates([(499, "offer-id-1", "id_1")])
        self.given_handler()
        self.when_increment_calls("offer-id-1")
        self.then_num_calls_are("offer-id-1", 500)

    def given_user_rates(self, rate_entry_values):
        rates = [create_rate_entry(*values_tuple) for values_tuple in rate_entry_values]
        try:
            self._user_rates_table.extend(rates)
        except AttributeError:
            self._user_rates_table = rates

    def given_handler(self):
        try:
            self._handler = MockHandlerUserRates(MockTableClient(self._user_rates_table))
        except AttributeError:
            self._handler = MockHandlerUserRates(MockTableClient([]))

    def when_increment_calls(self, offer_id):
        self._handler.increment_num_calls(offer_id)

    def then_num_calls_are(self, offer_id, num_calls):
        entities = self._handler.query_offer_id(offer_id)
        assert num_calls == entities[0][constants.NUM_CALLS]

    def test_rate_increased(self):
        self.given_user_rates([(500, "offer-id-2", "id-2")])
        self.given_handler()
        self.when_increment_calls("offer-id-2")
        self.then_num_calls_last_hour_is("offer-id-2", 1)

    def test_entry_updated_with_new_field(self):
        self.given_user_rates([(0, "offer-id-1", "id_1")])
        self.given_field_removed(constants.CALLS_LAST_PERIOD)
        self.given_handler()
        self.when_increment_calls("offer-id-1")
        self.then_num_calls_last_hour_is("offer-id-1", 1)

    def given_field_removed(self, field_name):
        for item in self._user_rates_table:
            del item[field_name]

    def then_num_calls_last_hour_is(self, offer_id, num_calls):
        entities = self._handler.query_offer_id_and_garbage_collect(offer_id)
        assert num_calls == len(entities[0][constants.CALLS_LAST_PERIOD])

    def test_old_entries_garbage_collected(self):
        self.given_user_rates([(500, "offer-id-2", "id-2")])
        self.given_old_call(61)
        self.given_handler()
        self.when_increment_calls("offer-id-2")
        self.then_num_calls_last_hour_is("offer-id-2", 1)

    def given_old_call(self, age_in_minutes):
        now = datetime.datetime.now()
        delta = delta = datetime.timedelta(minutes=age_in_minutes)
        birth_time = now - delta
        self._user_rates_table[0][constants.CALLS_LAST_PERIOD] = json.dumps([birth_time.timestamp()])

    def test_younger_entries_not_garbage_collected(self):
        self.given_user_rates([(500, "offer-id-2", "id-2")])
        self.given_old_call(59)
        self.given_handler()
        self.when_increment_calls("offer-id-2")
        self.then_num_calls_last_hour_is("offer-id-2", 2)

    def test_offer_not_in_db(self):
        self.given_handler()
        with pytest.raises(OfferNotInDB):
            self.when_increment_calls("new_offer")

    def test_create_entity(self):
        self.given_handler()
        self.when_create_entity("new_offer", "new_user", OFFER_USAGE_QUOTA)
        self.then_incrementing_works_flawlessly("new_offer")

    def when_create_entity(self, offer_id, user_id, offer_quota):
        self._handler.create_entity(offer_id, user_id, offer_quota)

    def then_incrementing_works_flawlessly(self, offer_id):
        self._handler.increment_num_calls(offer_id)
        assert True

    def test_create_entity_author_usage_quota(self):
        self.given_handler()
        self.when_create_entity_author_quota("new_user", AUTHOR_USAGE_QUOTA)
        self.then_user_rates_contain_author_entry("new_user")

    def when_create_entity_author_quota(self, author_user_id, author_quota):
        self._handler.create_entity_author_quota(author_user_id, author_quota)

    def then_user_rates_contain_author_entry(self, author_user_id):
        entities = self._handler.query_author_user_id(author_user_id)
        assert author_user_id == entities[0][constants.AUTHOR_ID]

    def test_update_author_quota(self):
        self.given_author_in_user_rates([("author-user-id-2", "id-2")])
        self.given_handler()
        self.when_update_author_quota("author-user-id-2", 4000)
        self.then_author_quota_is("author-user-id-2", 4000)

    def when_update_author_quota(self, author_user_id, author_quota):
        self._handler.update_author_quota(author_user_id, author_quota)

    def then_author_quota_is(self, author_user_id, author_quota):
        entities = self._handler.query_author_user_id(author_user_id)
        assert author_quota == entities[0][constants.AUTHOR_QUOTA]

    def given_author_in_user_rates(self, author_rates_values):
        author_rates = [create_author_quota_entry(*values_tuple) for values_tuple in author_rates_values]
        try:
            self._user_rates_table.extend(author_rates)
        except AttributeError:
            self._user_rates_table = author_rates


def create_buddy_generator_rate_entry(num_calls, author_user_id, row_key, author_quota=None):
    return {
        constants.PARTITION_KEY: constants.BUDDY_GENERATOR_DATA,
        constants.ROW_KEY: row_key,
        constants.NUM_CALLS: num_calls,
        constants.AUTHOR_ID: author_user_id,
        constants.AUTHOR_QUOTA: author_quota if author_quota else AUTHOR_GENERATION_QUOTA
    }


class TestBuddyGeneratorUserRates:
    def test_first_call(self):
        self.given_user_rates([(0, "author-user-id-2", "id-2")])
        self.given_handler()
        self.when_increment_calls("author-user-id-2")
        self.then_num_calls_are("author-user-id-2", 1)

    def given_user_rates(self, rate_entry_values):
        rates = [create_buddy_generator_rate_entry(*values_tuple) for values_tuple in rate_entry_values]
        try:
            self._user_rates_table.extend(rates)
        except AttributeError:
            self._user_rates_table = rates

    def given_handler(self):
        try:
            self._handler = MockBuddyGeneratorUserRates(MockTableClient(self._user_rates_table))
        except AttributeError:
            self._handler = MockBuddyGeneratorUserRates(MockTableClient([]))

    def when_increment_calls(self, author_user_id):
        self._handler.increment_num_calls(author_user_id)

    def then_num_calls_are(self, author_user_id, num_calls):
        entities = self._handler.query_author_user_id(author_user_id)
        assert num_calls == entities[0][constants.NUM_CALLS]

    def test_create_entity(self):
        self.given_handler()
        self.when_create_entity("new_user", AUTHOR_GENERATION_QUOTA)
        self.then_user_rates_contain_author_entry("new_user")

    def when_create_entity(self, author_user_id, author_quota):
        self._handler.create_entity(author_user_id, author_quota)

    def then_user_rates_contain_author_entry(self, author_user_id):
        entities = self._handler.query_author_user_id(author_user_id)
        assert author_user_id == entities[0][constants.AUTHOR_ID]

    def test_update_author_quota(self):
        self.given_user_rates([(0, "author-user-id-2", "id-2")])
        self.given_handler()
        self.when_update_author_quota("author-user-id-2", 40)
        self.then_author_quota_is("author-user-id-2", 40)

    def when_update_author_quota(self, author_user_id, author_quota):
        self._handler.update_author_quota(author_user_id, author_quota)

    def then_author_quota_is(self, author_user_id, author_quota):
        entities = self._handler.query_author_user_id(author_user_id)
        assert author_quota == entities[0][constants.AUTHOR_QUOTA]



import json

REQUEST_EVENT = "incoming request on frontend HTTP API"
RESPONSE_EVENT = "outgoing response on frontend HTTP API"
SELECTED_INTERPRETATION_EVENT = "DmeModule selected interpretation to act on"

BACKEND_VERSION_PREFIX = "Backend uses TDM version "
BACKEND_DIRECTORY_PREFIX = "Backend started in directory "
BACKEND_CONFIG_PREFIX = "Backend uses backend config "
BACKEND_LANGUAGE_PREFIX = "Backend uses language "


class UnexpectedPrefixException(Exception):
    pass


class LogExtractor:
    def __init__(self, logfile):
        self._logfile = logfile

    @property
    def _entries(self):
        with open(self._logfile) as file_:
            for line in file_:
                try:
                    yield json.loads(line)
                except ValueError:
                    continue

    def get_backend_communication_entries(self):
        for entry in self._entries:
            message = entry["event"]
            if message in [REQUEST_EVENT, RESPONSE_EVENT, SELECTED_INTERPRETATION_EVENT]:
                yield entry

    def get_backend_tdm_version(self):
        return self._get_first_suffix([BACKEND_VERSION_PREFIX])

    def get_backend_working_directory(self):
        return self._get_first_suffix([BACKEND_DIRECTORY_PREFIX])

    def get_backend_config(self):
        return self._get_first_suffix([BACKEND_CONFIG_PREFIX])

    def get_backend_language(self):
        return self._get_first_suffix([BACKEND_LANGUAGE_PREFIX])

    def _get_first_suffix(self, prefixes):
        suffixes = self._get_suffixes(prefixes)
        if suffixes:
            return suffixes[0]

    def _get_suffixes(self, prefixes):
        entries = self._entries_that_match(prefixes)
        suffixes = [self._extract_suffix_from_message(entry["event"], prefixes) for entry in entries]
        return suffixes

    def _entries_that_match(self, prefixes):
        def is_match(entry):
            return any([prefix in entry["event"] for prefix in prefixes])

        return [entry for entry in self._entries if is_match(entry)]

    def _extract_suffix_from_message(self, message, prefixes):
        start_of_prefix, prefix = self._find_start_of_a_prefix(message, prefixes)
        start_of_entry = start_of_prefix + len(prefix)
        suffix = message[start_of_entry:]
        return suffix

    def _find_start_of_a_prefix(self, message, prefixes):
        for prefix in prefixes:
            if prefix in message:
                start_of_entry = message.find(prefix)
                return start_of_entry, prefix
        raise UnexpectedPrefixException("Expected to find one of %s in %s but didn't" % (prefixes, message))

    @staticmethod
    def string_without_prefix(string, prefix):
        length_of_bytearray_prefix = len("b")
        start_of_content = len(prefix) + length_of_bytearray_prefix
        return string[start_of_content:]



from tala.ddd.files import ddd_files


class TestFiles:
    def setup_method(self):
        self._path = None
        self._result = None

    def test_ddd_files(self):
        self.given_path("ddds/mockup_travel")
        self.when_fetching_ddd_files()
        self.then_returns([
            'ddds/mockup_travel/__init__.py', 'ddds/mockup_travel/ddd.config.json', 'ddds/mockup_travel/domain.xml',
            'ddds/mockup_travel/http_service/http_service.py', 'ddds/mockup_travel/ontology.xml',
            'ddds/mockup_travel/service_interface.xml', 'ddds/mockup_travel/test/__init__.py',
            'ddds/mockup_travel/test/interaction_tests.json'
        ])

    def given_path(self, path):
        self._path = path

    def when_fetching_ddd_files(self):
        self._result = ddd_files(self._path)

    def then_returns(self, expected_result):
        actual_result = [str(actual) for actual in self._result]
        assert expected_result == sorted(actual_result), f"Expected {expected_result} but got {self._result}"



import copy
import warnings

from tala.model import move
from tala.model import plan_item
from tala.model import ontology
from tala.model.domain import Domain
from tala.model import condition
from tala.model.hint import Hint
from tala.model.action import Action
from tala.model.ask_feature import AskFeature
from tala.model.plan import Plan
from tala.model import sort
from tala.model.sort import Sort, CustomSort
from tala.model.predicate import Predicate
from tala.model.individual import No, Yes, Individual, NegativeIndividual
from tala.model.question import (
    Question, WhQuestion, YesNoQuestion, ConsequentQuestion, AltQuestion, KnowledgePreconditionQuestion
)
from tala.model.lambda_abstraction import (
    LambdaAbstractedProposition, LambdaAbstractedImplicationPropositionForConsequent, LambdaAbstractedGoalProposition,
    LambdaAbstractedPredicateProposition
)
from tala.model.proposition import (
    Proposition, GoalProposition, PropositionSet, PredicateProposition, ResolvednessProposition, PrereportProposition,
    ImplicationProposition, UnderstandingProposition, KnowledgePreconditionProposition, PreconfirmationProposition,
    RejectedPropositions, ServiceResultProposition, ActionStatusProposition, ServiceActionStartedProposition,
    ServiceActionTerminatedProposition, NumberOfAlternativesProposition, QuestionStatusProposition, MuteProposition,
    UnmuteProposition, QuitProposition, PredictedProposition
)
from tala.model.polarity import Polarity
from tala.model.goal import PerformGoal, ResolveGoal, PERFORM, RESOLVE
from tala.model.set import Set
from tala.model.openqueue import OpenQueue
from tala.model.date_time import DateTime
from tala.model.person_name import PersonName
from tala.model.service_action_outcome import SuccessfulServiceAction, FailedServiceAction
from tala.model.action_status import ActionStatus, Aborted, Done
from tala.model.ddd import DDD
from tala.ddd.services import service_interface
from tala.ddd.services.service_interface import ServiceInterface, ServiceActionInterface


class JSONParseFailure(Exception):
    pass


def preprocess_json_dict(json_dict_original):
    json_dict = copy.copy(json_dict_original)
    for float_key in ["perception_confidence", "understanding_confidence"]:
        if float_key in json_dict:
            try:
                json_dict[float_key] = float(json_dict[float_key])
            except TypeError:
                pass
    return json_dict


def create_sort(definition):
    sort_name = definition["_name"]
    if sort_name in sort.BUILTIN_SORTS:
        if sort_name == sort.DATETIME:
            return sort.DateTimeSort()
        if sort_name == sort.REAL:
            return sort.RealSort()
        if sort_name == sort.PERSON_NAME:
            return sort.PersonNameSort()
        if sort_name == sort.INTEGER:
            return sort.IntegerSort()
        if sort_name == sort.BOOLEAN:
            return sort.BooleanSort()
        if sort_name == sort.STRING:
            return sort.StringSort()
        if sort_name == sort.DOMAIN:
            return sort.DomainSort()
        if sort_name == sort.IMAGE:
            return sort.ImageSort()
        if sort_name == sort.WEBVIEW:
            return sort.WebviewSort()

    ontology_name = definition.get("_ontology_name")
    if ontology_name:
        return CustomSort(ontology_name, definition["_name"], definition["_dynamic"])
    return Sort(definition["_name"], definition["_dynamic"])


def create_predicate(name, definition):
    return Predicate(
        definition["_ontology_name"], name, create_sort(definition["sort"]), definition["feature_of_name"],
        definition["_multiple_instances"]
    )


def create_individual(individual, sort):
    def extract(string, prefix, suffix):
        if value.startswith(prefix) and value.endswith(suffix):
            return value[len(prefix):len(value) - len(suffix)]

    def extract_datetime_value():
        return extract(value, "datetime(", ")")

    def extract_person_name_value():
        return extract(value, "person_name(", ")")

    sort = create_sort(sort)
    value = individual
    if sort.is_datetime_sort():
        value = DateTime(extract_datetime_value())
    if sort.is_real_sort():
        value = float(value)
    if sort.is_person_name_sort():
        value = PersonName(extract_person_name_value())
    if sort.is_integer_sort():
        value = int(value)
    return (value, sort)


class JSONOntologyParser():
    def parse_ontology(self, ontology_as_json):
        if ontology_as_json["semantic_object_type"] == ontology.SEMANTIC_OBJECT_TYPE:
            name = ontology_as_json["_name"]
            sorts = self.parse_ontology_sorts(ontology_as_json["_sorts"])
            predicates = self.parse_ontology_predicates(ontology_as_json["_predicates"])
            individuals = self.parse_ontology_individuals(ontology_as_json["_individuals"])
            actions = self.parse_ontology_actions(ontology_as_json["_actions"], name)
            return ontology.Ontology(name, sorts, predicates, individuals, actions)
        raise JSONParseFailure(f"Failed to parse {ontology_as_json} as an ontology.")

    def parse_ontology_sorts(self, sorts_as_json):
        return set([create_sort(sort) for sort in sorts_as_json])

    def parse_ontology_predicates(self, predicates_as_json):
        return set([create_predicate(name, predicates_as_json[name]) for name in predicates_as_json])

    def parse_ontology_individuals(self, individuals_as_json):
        individuals = {}
        for individual_as_json in individuals_as_json:
            (value, sort) = create_individual(individual_as_json, individuals_as_json[individual_as_json])
            individuals[value] = sort
        return individuals

    def parse_ontology_actions(self, actions_as_json, ontology_name):
        return [action for action in actions_as_json["set"]]


class JSONDDDParser():
    def parse(self, ddd_as_json):
        name = ddd_as_json["ddd_name"]
        ontology_parser = JSONOntologyParser()
        ontology = ontology_parser.parse_ontology(ddd_as_json["ontology"])
        domain_parser = JSONDomainParser(ontology)
        domain = domain_parser.parse(ddd_as_json["domain"])
        service_interface_parser = JSONServiceInterfaceParser()
        service_interface = service_interface_parser.parse(ddd_as_json["service_interface"])
        return DDD(name, ontology, domain, service_interface)


class JSONDomainParser():
    def __init__(self, ontology=None):
        self.ontology = ontology
        self.parser = NonCheckingJSONParser()

    def parse(self, json_dict):
        ddd_name = json_dict["ddd_name"]
        domain_name = json_dict["name"]
        if self.ontology:
            self.parser = CheckingJSONParser(ddd_name, self.ontology, domain_name)
        default_questions = self.parse_default_questions(json_dict["default_questions"])
        parameters = self.parse_parameters(json_dict["parameters"])
        plans = self.parse_plans(json_dict["plans"])
        validators = self.parse_validators(json_dict["validators"])
        dependencies = self.parse_serialized_dict(json_dict.get("dependencies", {}))
        queries = self.parse_serialized_query_list(json_dict["queries"])
        return Domain(
            ddd_name, domain_name, self.ontology, plans, default_questions, parameters, validators, dependencies,
            queries
        )

    def parse_default_questions(self, default_questions):
        return [self._parse_default_question(question) for question in default_questions]

    def _parse_default_question(self, question):
        return self.parser.parse(question)

    def parse_parameters(self, parameters):
        compiled_parameters = {}
        for pair in parameters:
            key_question = self.parser.parse(pair["key"])
            values = pair["value"]
            compiled_parameters[key_question] = self._parse_parameters_for_entry(values)
        return compiled_parameters

    def _parse_parameters_for_entry(self, parameters):
        return {name: self._parse_parameter(parameters, name) for name in parameters}

    def _parse_parameter(self, parameters, name):
        def is_questions_parameter():
            return name in ["related_information"]

        def is_question_parameter():
            return name in ["service_query", "default"]

        def is_int_parameter():
            return name in ["max_spoken_alts", "max_reported_hit_count"]

        def is_boolean_parameter():
            return name in ["always_ground", "always_relevant"]

        if name == "ask_features":
            return [self.parse_ask_feature(feature) for feature in parameters[name]]
        if name == "hints":
            return [self.parse_hint(hint) for hint in parameters[name]]
        if name in ["allow_goal_accommodation", "verbalize", "incremental"]:
            return parameters[name]
        if name in ["on_zero_hits_action", "on_too_many_hits_action"]:
            return self.parser.parse_action(parameters[name])
        if is_int_parameter():
            return int(parameters[name])
        if name in ["source"]:
            return parameters[name]
        if name == "alts":
            return PropositionSet([self.parser.parse(proposition) for proposition in parameters[name]["_propositions"]])
        if name == "background":
            return [self.parser.parse_predicate(predicate) for predicate in parameters[name]]
        if is_questions_parameter():
            return [self.parser.parse_question(question) for question in parameters[name]]
        if is_question_parameter():
            return self.parser.parse_question(parameters[name])
        if is_boolean_parameter():
            return bool(parameters[name])

        raise JSONParseFailure("no support for parameter", name, parameters[name], ".")
        return parameters[name]

    def parse_ask_feature(self, feature):
        return AskFeature(feature["name"], feature["kpq"])

    def parse_hint(self, hint):
        return Hint([self.parser.parse_plan_item(item) for item in hint["_plan"]])

    def parse_plans(self, plans):
        def is_builtin(goal):
            return goal["_goal_type"] == "PERFORM_GOAL" and goal["_content"]["value"] in ["top", "up"]

        return [self.parse_goal(plans[goal]) for goal in plans]

    def parse_goal(self, goal):
        result = {}
        BOOLEAN_VALUES = [
            "unrestricted_accommodation", "restart_on_completion", "reraise_on_resume", "accommodate_without_feedback"
        ]
        INTEGER_VALUES = ["max_answers"]
        STRING_VALUES = ["alternatives_predicate"]
        for entry in goal:
            if entry == "goal":
                result[entry] = self.parser.parse_goal(goal[entry])
            elif entry == "io_status":
                warnings.warn("io_status is deprecated.", DeprecationWarning, stacklevel=2)
                result[entry] = self.parse_io_status(goal[entry])
            elif entry == "plan":
                result[entry] = self.parse_plan(goal[entry])
            elif entry == "postconds":
                result[entry] = self.parse_postconds(goal[entry])
            elif entry == "postplan":
                result[entry] = self.parse_postplan(goal[entry])
            elif entry == "preferred":
                result[entry] = self.parse_preferred(goal[entry])
            elif entry == "superactions":
                result[entry] = self.parse_superactions(goal[entry])
            elif entry in BOOLEAN_VALUES:
                result[entry] = bool(goal[entry])
            elif entry in INTEGER_VALUES:
                result[entry] = int(goal[entry])
            elif entry in STRING_VALUES:
                result[entry] = str(goal[entry])
            else:
                raise JSONParseFailure("found unknown entry in goal when parsing:", entry, goal[entry])

        return result

    def parse_plan(self, plan):
        plan_items = [self.parser.parse_plan_item(item) for item in plan["stack"]]
        return Plan(reversed(plan_items))

    def parse_io_status(self, status):
        if status:
            return status
        return Domain.DEFAULT_IO_STATUS

    def parse_postconds(self, postconds):
        return [self.parse_postcond(postcond) for postcond in postconds]

    def parse_postcond(self, postcond):
        return self.parser.parse(postcond)

    def parse_postplan(self, postplan):
        try:
            return [self.parser.parse_plan_item(item) for item in postplan["stack"]]
        except TypeError:
            return [self.parser.parse_plan_item(item) for item in postplan]

    def parse_superactions(self, actions):
        return [self.parse_superaction(action) for action in actions]

    def parse_preferred(self, preference):
        try:
            return self.parser.parse_proposition(preference)
        except TypeError:
            return preference

    def parse_superaction(self, action):
        return self.parser.parse_action(action)

    def parse_validators(self, validators):
        return [self.parse_validator(validator) for validator in validators]

    def parse_validator(self, validator):
        def parse_propositions_list(predicates):
            return [self.parser.parse(predicate) for predicate in predicates]

        id = validator["validator"]
        valid_configurations = [
            parse_propositions_list(configuration) for configuration in validator["valid_configurations"]
        ]
        return {"validator": id, "valid_configurations": valid_configurations}

    def parse_serialized_dict(self, serialized_dict):
        result_dict = {}
        for pair in serialized_dict:
            key = self.parser.parse(pair["key"])
            value = pair["value"]
            result_dict[key] = self.parser.parse(value)
        return result_dict

    def parse_serialized_query_list(self, serialized_queries):
        def is_iterator(query_content):
            return "limit" in query_content

        def parse_iterator(iterator):
            result_dict = {}
            for key, value in iterator.items():
                if key in ["for_enumeration", "for_random_enumeration"]:
                    result_dict[key] = [self.parser.parse(item) for item in value]
                else:
                    result_dict[key] = value
            return result_dict

        result_list = []
        for query in serialized_queries:
            if is_iterator(query):
                query_dict = parse_iterator(query)
            else:
                query_dict = {}
                for entry, semantic_objects in query.items():
                    try:
                        query_dict[entry] = self.parser.parse(semantic_objects)
                    except AttributeError:
                        query_dict[entry] = [self.parser.parse(item) for item in semantic_objects]
            result_list.append(query_dict)
        return result_list


class JSONServiceInterfaceParser():
    def parse(self, json_dict):
        actions = self.parse_actions(json_dict["_actions"])
        queries = self.parse_queries(json_dict["_queries"])
        validators = self.parse_validators(json_dict["_validators"])
        return ServiceInterface(actions, queries, validators)

    def parse_actions(self, actions_dict):
        return [self.parse_action_interface(actions_dict[action_name]) for action_name in actions_dict]

    def parse_action_interface(self, action_dict):
        name = action_dict["_name"]
        target = self.parse_target(action_dict["_target"])
        service_parameters = self.parse_parameters(action_dict["_parameters"])
        failure_reasons = action_dict["_failure_reasons"]
        interface = ServiceActionInterface(name, target, service_parameters, failure_reasons)
        return interface

    def parse_target(self, target_dict):
        if target_dict.get("target_type") == service_interface.FRONTEND_TARGET:
            return service_interface.FrontendTarget()
        if target_dict.get("target_type") == service_interface.HTTP_TARGET:
            endpoint = target_dict["_endpoint"]
            return service_interface.HttpTarget(endpoint)
        raise JSONParseFailure(f"Cannot parse as target: {target_dict}")

    def parse_parameters(self, parameter_dict):
        return [self.parse_parameter(parameter) for parameter in parameter_dict]

    def parse_parameter(self, parameter):
        name = parameter["_name"]
        format = parameter["_format"]
        is_optional = parameter["_is_optional"]
        return service_interface.ServiceParameter(name, format, is_optional)

    def parse_queries(self, queries_dict):
        return [self.parse_query(query) for query in queries_dict.values()]

    def parse_query(self, query_dict):
        name = query_dict["_name"]
        target = self.parse_target(query_dict["_target"])
        parameters = self.parse_parameters(query_dict["_parameters"])
        return service_interface.ServiceQueryInterface(name, target, parameters)

    def parse_validators(self, validators_dict):
        return [self.parse_validator(validator) for validator in validators_dict.values()]

    def parse_validator(self, validator_dict):
        name = validator_dict["_name"]
        target = self.parse_target(validator_dict["_target"])
        parameters = self.parse_parameters(validator_dict["_parameters"])
        return service_interface.ServiceValidatorInterface(name, target, parameters)


class NonCheckingJSONParser():
    def parse(self, json_dict_unchecked):
        def is_condition():
            """ This is the template for extending the semantic object JSON
            with type information, which should be done for all semantic objects.
            """

            return json_dict.get("semantic_object_type") == condition.SEMANTIC_OBJECT_TYPE

        def is_plan_item():
            for item in plan_item.ALL_PLAN_ITEM_TYPES:
                if item in json_dict:
                    return True
            return False

        def is_individual():
            return "value" in json_dict and "sort" in json_dict

        def is_yes_no():
            return json_dict.get("semantic_object_type") == "yes/no"

        def is_move():
            return "move_type" in json_dict

        def is_action():
            return "value" in json_dict

        def is_goal():
            return "_goal_type" in json_dict

        def is_question():
            question_type = json_dict.get("_type", None)
            return question_type in Question.TYPES

        def is_predicate():
            return "name" in json_dict and "sort" in json_dict

        def is_lambda_abstracted_proposition():
            return json_dict.get("semantic_object_type") == "LambdaAbstractedProposition"

        def is_set():
            return "set" in json_dict

        def is_openqueue():
            return "openqueue" in json_dict

        def is_service_action_outcome():
            return "service_action_outcome" in json_dict

        def is_proposition():
            prop_type = json_dict.get("_type", None)
            return prop_type in Proposition.TYPES

        def is_proposition_set():
            return "_propositions" in json_dict

        def is_action_status():
            return "action_status" in json_dict

        json_dict = preprocess_json_dict(json_dict_unchecked)
        if is_condition():
            return self.parse_condition(json_dict)
        if is_plan_item():
            return self.parse_plan_item(json_dict)
        if is_individual():
            return self.parse_individual(json_dict)
        if is_set():
            return self.parse_set(json_dict["set"])
        if is_openqueue():
            return self.parse_openqueue(json_dict["openqueue"])
        if is_move():
            return self.parse_move(json_dict)
        if is_action():
            return self.parse_action(json_dict)
        if is_goal():
            return self.parse_goal(json_dict)
        if is_question():
            return self.parse_question(json_dict)
        if is_predicate():
            return self.parse_predicate(json_dict)
        if is_proposition_set():
            return self.parse_proposition_set(json_dict)
        if is_proposition():
            return self.parse_proposition(json_dict)
        if is_lambda_abstracted_proposition():
            return self.parse_lambda_abstraction(json_dict)
        if is_service_action_outcome():
            return self.parse_service_action_outcome(json_dict)
        if is_yes_no():
            if json_dict["instance"] == No.NO:
                return No()
            if json_dict["instance"] == Yes.YES:
                return Yes()
        if is_action_status():
            return self.parse_action_status(json_dict)
        raise JSONParseFailure(f"JSONParser cannot parse: {json_dict}")

    def parse_condition(self, condition_json):
        keys = ["_predicate", "_proposition"]
        for key in keys:
            if condition_json.get(key):
                argument = self.parse(condition_json[key])
                return condition.create_condition(condition_json["type"], argument)
        if condition_json.get("_query"):
            try:
                argument = self.parse(condition_json["_query"])
            except Exception:
                argument = condition_json["_query"]
            return condition.create_condition(condition_json["type"], argument)
        raise JSONParseFailure(f"Failed to parse {condition_json} as a condition.")

    def parse_plan_item(self, data):
        if plan_item.TYPE_FINDOUT in data:
            return self.parse_findout_plan_item(data)
        if plan_item.TYPE_RAISE in data:
            return self.parse_raise_plan_item(data)
        if plan_item.TYPE_RESPOND in data:
            return self.parse_respond_plan_item(data[plan_item.TYPE_RESPOND])
        if plan_item.TYPE_ASSUME in data:
            return self.parse_assume_plan_item(data[plan_item.TYPE_ASSUME])
        if plan_item.TYPE_ASSUME_ISSUE in data:
            return self.parse_assume_issue_plan_item(data)
        if plan_item.TYPE_ASSUME_SHARED in data:
            return self.parse_assume_shared_plan_item(data[plan_item.TYPE_ASSUME_SHARED])
        if plan_item.TYPE_BIND in data:
            return self.parse_bind_plan_item(data[plan_item.TYPE_BIND])
        if plan_item.TYPE_DO in data:
            return self.parse_do_plan_item(data[plan_item.TYPE_DO])
        if plan_item.TYPE_EMIT_ICM in data:
            return self.parse_emit_icm_plan_item(data[plan_item.TYPE_EMIT_ICM])
        if plan_item.TYPE_FORGET in data:
            return self.parse_forget_plan_item(data[plan_item.TYPE_FORGET])
        if plan_item.TYPE_FORGET_SHARED in data:
            return self.parse_forget_shared_plan_item(data[plan_item.TYPE_FORGET_SHARED])
        if plan_item.TYPE_FORGET_ALL in data:
            return self.parse_forget_all_plan_item(data[plan_item.TYPE_FORGET_ALL])
        if plan_item.TYPE_FORGET_ISSUE in data:
            return self.parse_forget_issue_plan_item(data[plan_item.TYPE_FORGET_ISSUE])
        if plan_item.TYPE_IF_THEN_ELSE in data:
            return self.parse_if_then_else_plan_item(data[plan_item.TYPE_IF_THEN_ELSE])
        if plan_item.TYPE_JUMPTO in data:
            return self.parse_jumpto_plan_item(data[plan_item.TYPE_JUMPTO])
        if plan_item.TYPE_INVOKE_SERVICE_QUERY in data:
            return self.parse_invoke_service_query_plan_item(data[plan_item.TYPE_INVOKE_SERVICE_QUERY])
        if plan_item.TYPE_INVOKE_DOMAIN_QUERY in data:
            return self.parse_invoke_domain_query_plan_item(data[plan_item.TYPE_INVOKE_DOMAIN_QUERY])
        if plan_item.TYPE_INVOKE_SERVICE_ACTION in data:
            return self.parse_invoke_service_action_plan_item(data[plan_item.TYPE_INVOKE_SERVICE_ACTION])
        if plan_item.TYPE_LOG in data:
            return self.parse_log_plan_item(data[plan_item.TYPE_LOG])
        if plan_item.TYPE_ACTION_PERFORMED in data:
            return self.parse_action_performed_plan_item(data[plan_item.TYPE_ACTION_PERFORMED])
        if plan_item.TYPE_ACTION_ABORTED in data:
            return self.parse_action_aborted_plan_item(data[plan_item.TYPE_ACTION_ABORTED])
        if plan_item.TYPE_CHANGE_DDD in data:
            return self.parse_change_ddd_plan_item(data[plan_item.TYPE_CHANGE_DDD])
        if plan_item.TYPE_GET_DONE in data:
            return self.parse_get_done_plan_item(data)
        if plan_item.TYPE_INVOKE_DOMAIN_QUERY in data:
            return self.parse_invoke_domain_query(data[plan_item.INVOKE_DOMAIN_QUERY])
        if plan_item.TYPE_END_TURN in data:
            return self.parse_end_turn_plan_item(data[plan_item.TYPE_END_TURN])
        if plan_item.TYPE_RESET_DOMAIN_QUERY in data:
            return self.parse_reset_domain_query(data[plan_item.TYPE_RESET_DOMAIN_QUERY])
        if plan_item.TYPE_ITERATE in data:
            return self.parse_iterate_plan_item(data[plan_item.TYPE_ITERATE])
        if plan_item.TYPE_ACTION_REPORT in data:
            return self.parse_action_report_plan_item(data[plan_item.TYPE_ACTION_REPORT])
        if plan_item.TYPE_GREET in data:
            return plan_item.Greet()
        if plan_item.TYPE_EMIT_MOVE in data:
            return self.parse_emit_move_plan_item(data)
        if plan_item.TYPE_QUESTION_REPORT in data:
            return self.parse_question_report_plan_item(data)
        if plan_item.TYPE_SERVICE_REPORT in data:
            return self.parse_service_report_plan_item(data)
        if plan_item.TYPE_RESPOND_TO_INSULT in data:
            return plan_item.RespondToInsult()
        if plan_item.TYPE_RESPOND_TO_THANK_YOU in data:
            return plan_item.RespondToThankYou()

        raise JSONParseFailure(f"PlanItem {data} not supported by json parser")

    def parse_findout_plan_item(self, data):
        allow_answer_from_pcom = data.get("allow_answer_from_pcom", False)
        domain_name = data.get("domain_name")
        question = self.parse_question(data[plan_item.TYPE_FINDOUT])
        return plan_item.Findout(domain_name, question, allow_answer_from_pcom)

    def parse_raise_plan_item(self, data):
        domain_name = data.get("domain_name")
        question = self.parse_question(data[plan_item.TYPE_RAISE])
        return plan_item.Raise(domain_name, question)

    def parse_respond_plan_item(self, data):
        question = self.parse_question(data)
        return plan_item.Respond(question)

    def parse_bind_plan_item(self, data):
        question = self.parse_question(data)
        return plan_item.Bind(question)

    def parse_do_plan_item(self, data):
        action = self.parse_action(data)
        return plan_item.Do(action)

    def parse_log_plan_item(self, data):
        if "message" in data and "level" in data:
            return plan_item.Log(data["message"], data["level"])
        else:
            return self.parse_old_log_plan_item(data)

    def parse_old_log_plan_item(self, data):
        message = data
        return plan_item.Log(message)

    def parse_action_performed_plan_item(self, data):
        return plan_item.GoalPerformed(data)

    def parse_action_aborted_plan_item(self, data):
        return plan_item.GoalAborted(data)

    def parse_change_ddd_plan_item(self, data):
        return plan_item.ChangeDDD(data)

    def parse_get_done_plan_item(self, data):
        action = self.parse_action(data[plan_item.TYPE_GET_DONE])
        try:
            step = int(data.get("step"))
        except TypeError:
            step = None

        return plan_item.GetDone(action, step)

    def parse_end_turn_plan_item(self, data):
        timeout = float(data)
        return plan_item.EndTurn(timeout)

    def parse_reset_domain_query(self, data):
        query = self.parse_question(data)
        return plan_item.ResetDomainQuery(query)

    def parse_iterate_plan_item(self, iterator_name):
        return plan_item.Iterate(iterator_name)

    def parse_action_report_plan_item(self, result_proposition):
        return plan_item.ActionReport(self.parse_proposition(result_proposition))

    def parse_emit_move_plan_item(self, data):
        prereport_move = self.parse_prereport_move(data[plan_item.TYPE_EMIT_MOVE])
        return plan_item.EmitMove(prereport_move)

    def parse_invoke_domain_query(self, data):
        issue = self.parse_question(data)
        return plan_item.InvokeDomainQuery(issue)

    def parse_forget_plan_item(self, data):
        if "_type" in data.keys():
            proposition = self.parse_proposition(data)
            return plan_item.Forget(proposition)
        else:
            predicate = self.parse_predicate(data)
            return plan_item.Forget(predicate)

    def parse_forget_shared_plan_item(self, data):
        if "_type" in data.keys():
            proposition = self.parse_proposition(data)
            return plan_item.ForgetShared(proposition)
        else:
            predicate = self.parse_predicate(data)
            return plan_item.ForgetShared(predicate)

    def parse_forget_all_plan_item(self, _data):
        return plan_item.ForgetAll()

    def parse_forget_issue_plan_item(self, data):
        question = self.parse_question(data)
        return plan_item.ForgetIssue(question)

    def parse_assume_plan_item(self, data):
        proposition = self.parse_proposition(data)
        return plan_item.Assume(proposition)

    def parse_assume_issue_plan_item(self, data):
        question = self.parse_question(data[plan_item.TYPE_ASSUME_ISSUE])
        return plan_item.AssumeIssue(question, data["insist"])

    def parse_assume_shared_plan_item(self, data):
        proposition = self.parse_proposition(data)
        return plan_item.AssumeShared(proposition)

    def parse_emit_icm_plan_item(self, data):
        icm = self.parse_move(data)
        return plan_item.EmitIcm(icm)

    def parse_jumpto_plan_item(self, data):
        goal = self.parse_goal(data)
        return plan_item.JumpTo(goal)

    def parse_if_then_else_plan_item(self, data):
        cond = self.parse(data["condition"])
        consequent = [self.parse(element) for element in data["consequent"]]
        alternative = [self.parse(element) for element in data["alternative"]]
        return plan_item.IfThenElse(cond, consequent, alternative)

    def parse_invoke_service_query_plan_item(self, data):
        issue = self.parse_question(data["issue"])
        min_results = int(data["min_results"])
        try:
            max_results = int(data["max_results"])
        except TypeError:
            max_results = None
        return plan_item.InvokeServiceQuery(issue, min_results=min_results, max_results=max_results)

    def parse_invoke_domain_query_plan_item(self, data):
        issue = self.parse_question(data["issue"])
        min_results = int(data["min_results"])
        max_results = int(data["max_results"])
        return plan_item.InvokeDomainQuery(issue, min_results=min_results, max_results=max_results)

    def parse_invoke_service_action_plan_item(self, data):
        service_action = data["service_action"]
        preconfirm = data["preconfirm"]
        postconfirm = data["postconfirm"]
        downdate_plan = data["downdate_plan"]
        ontology_name = data["ontology"]
        return plan_item.InvokeServiceAction(ontology_name, service_action, preconfirm, postconfirm, downdate_plan)

    def parse_question_report_plan_item(self, data):
        result_proposition = self.parse_proposition(data[plan_item.TYPE_QUESTION_REPORT])
        return plan_item.QuestionReport(result_proposition)

    def parse_service_report_plan_item(self, data):
        result_proposition = self.parse_proposition(data[plan_item.TYPE_SERVICE_REPORT])
        return plan_item.ServiceReport(result_proposition)

    def parse_question(self, data):
        if data["_type"] == Question.TYPE_WH:
            lambda_abstraction = self.parse_lambda_abstraction(data["_content"])
            return WhQuestion(lambda_abstraction)
        if data["_type"] == Question.TYPE_KPQ:
            embedded_question = self.parse_question(data["_content"])
            return KnowledgePreconditionQuestion(embedded_question)
        if data["_type"] == Question.TYPE_YESNO:
            proposition = self.parse(data["_content"])
            return YesNoQuestion(proposition)
        if data["_type"] == Question.TYPE_ALT:
            p_set = self.parse_proposition_set(data["_content"])
            return AltQuestion(p_set)
        if data["_type"] == Question.TYPE_CONSEQUENT:
            lambda_abstraction = self.parse_lambda_abstraction(data["_content"])
            return ConsequentQuestion(lambda_abstraction)
        raise JSONParseFailure(f"Cannot parse {data} as question.")

    def parse_lambda_abstraction(self, content):
        if content["type_"] == LambdaAbstractedProposition.LAMBDA_ABSTRACTED_PREDICATE_PROPOSITION:
            return self._create_lambda_abstracted_predicate_proposition(content)

        if content["type_"] == LambdaAbstractedProposition.LAMBDA_ABSTRACTED_GOAL_PROPOSITION:
            return LambdaAbstractedGoalProposition()

        if content["type_"] == LambdaAbstractedProposition.LAMBDA_ABSTRACTED_IMPLICATION_PROPOSITION_FOR_CONSEQUENT:
            ontology_name = content["_ontology_name"]
            antecedent = self.parse_proposition(content["_antecedent"])
            predicate = self.parse_predicate(content["_consequent_predicate"])
            return LambdaAbstractedImplicationPropositionForConsequent(antecedent, predicate, ontology_name)
        raise JSONParseFailure(f"Cannot parse {content} as lambda abstraction.")

    def _create_lambda_abstracted_predicate_proposition(self, content):
        predicate = self.parse_predicate(content["predicate"])
        ontology_name = content["_ontology_name"]
        return LambdaAbstractedPredicateProposition(predicate, ontology_name)

    def parse_predicate(self, data):
        return create_predicate(data["name"], data)

    def parse_proposition_set(self, data):
        proposition_list = data["_propositions"]
        propositions = [self.parse_proposition(proposition) for proposition in proposition_list]
        polarity = data["_polarity"]
        return PropositionSet(propositions, polarity)

    def parse_set(self, data):
        result_set = Set([self.parse(element) for element in data])
        return result_set

    def parse_openqueue(self, input):
        front_content = [self.parse(item) for item in input["front_content"]]
        back_content = [self.parse(item) for item in input["back_content"]]
        unshifted_as_json = input["unshifted_content"]
        if unshifted_as_json:
            unshifted_content = [self.parse(item) for item in unshifted_as_json]
        else:
            unshifted_content = None
        return OpenQueue(front_content, back_content, unshifted_content)

    def parse_move(self, data):
        def get_float(string):
            if string is not None:
                return float(string)
            return None

        move_args = {
            "understanding_confidence": get_float(data.get("understanding_confidence")),
            "speaker": data.get("speaker"),
            "ddd_name": data.get("ddd"),
            "perception_confidence": get_float(data.get("perception_confidence"))
        }
        non_icm_args = {
            "modality": data.get("modality"),
            "weighted_understanding_confidence": get_float(data.get("weighted_understanding_confidence")),
            "utterance": data.get("utterance")
        } | move_args

        icm_args = {
            "polarity": data.get("polarity"),
        } | move_args

        if data["move_type"] == move.Move.ANSWER:
            return self.parse_answer_move(data, non_icm_args)
        if data["move_type"] == move.Move.REQUEST:
            return self.parse_request_move(data, non_icm_args)
        if data["move_type"] == move.Move.ASK:
            return self.parse_ask_move(data, non_icm_args)
        if data["move_type"] == move.ICMMove.ACC:
            return self.parse_acc_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.ACCOMMODATE:
            return self.parse_accommodate_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.RESUME:
            return self.parse_resume_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.RERAISE:
            return self.parse_reraise_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.PER:
            return self.parse_per_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.UND:
            return self.parse_und_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.SEM:
            return self.parse_sem_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.LOADPLAN:
            return self.parse_loadplan_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.CARDINAL_SEQUENCING:
            return self.parse_cardinal_sequencing_icm(data, icm_args)
        if data["move_type"] == move.ICMMove.REPORT_INFERENCE:
            return self.parse_report_inference_icm(data, icm_args)
        if data["move_type"] == move.Move.GREET:
            return self.parse_greet_move(non_icm_args)
        if data["move_type"] == move.Move.INSULT:
            return self.parse_insult_move(non_icm_args)
        if data["move_type"] == move.Move.INSULT_RESPONSE:
            return self.parse_insult_response_move(non_icm_args)
        if data["move_type"] == move.Move.MUTE:
            return self.parse_mute_move(non_icm_args)
        if data["move_type"] == move.Move.UNMUTE:
            return self.parse_unmute_move(non_icm_args)
        if data["move_type"] == move.Move.PREREPORT:
            return self.parse_prereport_move(data)
        if data["move_type"] == move.Move.REPORT:
            return self.parse_report_move(data, non_icm_args)
        if data["move_type"] == move.Move.QUIT:
            return self.parse_quit_move(non_icm_args)
        if data["move_type"] == move.Move.THANK_YOU:
            return self.parse_thank_you_move(non_icm_args)
        if data["move_type"] == move.Move.THANK_YOU_RESPONSE:
            return self.parse_thank_you_response_move(non_icm_args)
        raise JSONParseFailure(f"cannot parse {data} as move.")

    def parse_action(self, data):
        action_name = data["value"]
        ontology_name = data["_ontology_name"]
        return Action(action_name, ontology_name)

    def parse_answer_move(self, data, kwargs):
        content = self.parse(data["content"])
        return move.AnswerMove(content, **kwargs)

    def parse_request_move(self, data, kwargs):
        action = self.parse_action(data["content"])
        return move.RequestMove(action, **kwargs)

    def parse_proposition(self, data):
        if data["_type"] == Proposition.PREDICATE:
            return self._create_predicate_proposition(data)
        if data["_type"] == Proposition.GOAL:
            goal = self.parse_goal(data["_content"])
            polarity = data.get("_polarity")
            return GoalProposition(goal, polarity)
        if data["_type"] == Proposition.PRECONFIRMATION:
            service_action = data.get("service_action")
            polarity = data.get("polarity")
            arguments = [self.parse_proposition(proposition) for proposition in data.get("_arguments")]
            ontology_name = data["_ontology_name"]
            return PreconfirmationProposition(ontology_name, service_action, arguments, polarity)
        if data["_type"] == Proposition.SERVICE_RESULT:
            service_action = data.get("service_action")
            action_outcome = self.parse(data.get("result"))
            arguments = [self.parse_proposition(proposition) for proposition in data.get("arguments")]
            ontology_name = data.get("ontology_name")
            return ServiceResultProposition(ontology_name, service_action, arguments, action_outcome)
        if data["_type"] == Proposition.UNDERSTANDING:
            content = self.parse(data["_content"])
            polarity = data.get("_polarity")
            speaker = data.get("_speaker")
            return UnderstandingProposition(speaker, content, polarity)
        if data["_type"] == Proposition.RESOLVEDNESS:
            issue = self.parse_question(data["issue"])
            return ResolvednessProposition(issue)
        if data["_type"] == Proposition.KNOWLEDGE_PRECONDITION:
            question = self.parse_question(data["_content"])
            polarity = data.get("_polarity")
            return KnowledgePreconditionProposition(question, polarity)
        if data["_type"] == Proposition.ACTION_STATUS:
            status = self.parse_action_status(data["_status"])
            action = self.parse_action(data["_content"])
            return ActionStatusProposition(action, status)
        if data["_type"] == Proposition.SERVICE_ACTION_STARTED:
            service_action = data["service_action"]
            parameters = [self.parse_proposition(proposition) for proposition in data.get("parameters")]
            ontology_name = data.get("ontology_name")
            return ServiceActionStartedProposition(ontology_name, service_action, parameters)
        if data["_type"] == Proposition.SERVICE_ACTION_TERMINATED:
            service_action = data["service_action"]
            polarity = data.get("_polarity")
            ontology_name = data["_ontology_name"]
            return ServiceActionTerminatedProposition(ontology_name, service_action, polarity)
        if data["_type"] == Proposition.PREREPORT:
            service_action = data["service_action"]
            argument_list = [self.parse_proposition(proposition) for proposition in data["argument_set"]]
            ontology_name = data["_ontology_name"]
            return PrereportProposition(ontology_name, service_action, argument_list)
        if data["_type"] == Proposition.REJECTED:
            polarity = data.get("_polarity")
            reason_for_rejection = data.get("reason_for_rejection")
            combination = self.parse(data["rejected_combination"])
            return RejectedPropositions(combination, polarity, reason_for_rejection)
        if data["_type"] == Proposition.IMPLICATION:
            antecedent = self.parse_proposition(data["_antecedent"])
            consequent = self.parse_proposition(data["_consequent"])
            return ImplicationProposition(antecedent, consequent)
        if data["_type"] == Proposition.NUMBER_OF_ALTERNATIVES:
            individual = self.parse_individual(data["_content"])
            return NumberOfAlternativesProposition(individual)
        if data["_type"] == Proposition.QUESTION_STATUS:
            question = self.parse_question(data["_content"])
            status = self.parse_action_status(data["_status"])
            return QuestionStatusProposition(question, status)
        if data["_type"] == Proposition.MUTE:
            return MuteProposition()
        if data["_type"] == Proposition.UNMUTE:
            return UnmuteProposition()
        if data["_type"] == Proposition.QUIT:
            return QuitProposition()
        if data["_type"] == Proposition.PREDICTED:
            return self._create_predicted_proposition(data)
        raise JSONParseFailure(f"Cannot parse {data} as proposition.")

    def _create_predicate_proposition(self, data):
        if data["predicate"]["sort"]["_name"] == "boolean":
            individual = None
        elif data["predicate"]["sort"]["_name"] == "string" and data["individual"] is None:
            individual = None
        else:
            individual = self.parse_individual(data["individual"])
        predicate = self.parse_predicate(data["predicate"])
        polarity = data.get("_polarity")
        return PredicateProposition(predicate, individual, polarity)

    def _create_predicted_proposition(self, data):
        prediction = self.parse_proposition(data["_content"])
        polarity = data.get("_polarity")
        return PredictedProposition(prediction, polarity)

    def parse_individual(self, data):
        if data is None:
            return None
        sort_data = data["sort"]
        value_data = data["value"]
        value, sort = create_individual(value_data, sort_data)
        polarity = data["polarity"]
        ontology_name = data["_ontology_name"]
        if polarity == Polarity.POS:
            return Individual(ontology_name, value, sort)
        return NegativeIndividual(ontology_name, value, sort)

    def parse_goal(self, data):
        if data is None:
            return None
        if data.get("_goal_type") == PERFORM:
            target = data.get("_target")
            action = self.parse_action(data["_content"])
            goal = PerformGoal(action, target=target)
            return goal
        if data.get("_goal_type") == RESOLVE:
            target = data.get("_target")
            question = self.parse_question(data["_content"])
            return ResolveGoal(question, target=target)
        raise JSONParseFailure(f"could not parse {data} as goal.")

    def parse_acc_icm(self, data, kwargs):
        if data.get("content"):
            content = self.parse(data["content"])
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.ACC, content, content_speaker=content_speaker, **kwargs)

        return move.ICMMove(move.ICMMove.ACC, **kwargs)

    def parse_accommodate_icm(self, data, kwargs):
        if data.get("content"):
            content = self.parse(data["content"])
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.ACCOMMODATE, content, content_speaker=content_speaker, **kwargs)
        else:
            return move.ICMMove(move.ICMMove.ACCOMMODATE, **kwargs)

    def parse_resume_icm(self, data, kwargs):
        goal = self.parse_goal(data["content"])
        content_speaker = data.get("content_speaker")
        return move.ICMMoveWithContent(move.ICMMove.RESUME, goal, content_speaker=content_speaker, **kwargs)

    def parse_reraise_icm(self, data, kwargs):
        if data.get("content"):
            action = self.parse(data["content"])
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.RERAISE, action, content_speaker=content_speaker, **kwargs)
        else:
            return move.ICMMove(move.ICMMove.RERAISE, **kwargs)

    def parse_per_icm(self, data, kwargs):
        string = data.get("content", None)
        if string:
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithContent(move.ICMMove.PER, string, content_speaker=content_speaker, **kwargs)
        return move.ICMMove(move.ICMMove.PER, **kwargs)

    def parse_und_icm(self, data, kwargs):
        content_data = data.get("content")
        if content_data:
            content = self.parse(content_data)
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithSemanticContent(move.ICMMove.UND, content, content_speaker=content_speaker, **kwargs)
        else:
            return move.ICMMove(move.ICMMove.UND, **kwargs)

    def parse_sem_icm(self, data, kwargs):
        content_data = data.get("content")
        if content_data:
            content = self.parse_proposition(content_data)
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithSemanticContent(move.ICMMove.SEM, content, content_speaker=content_speaker, **kwargs)
        content_speaker = data.get("content_speaker")
        return move.ICMMove(move.ICMMove.SEM, **kwargs)

    def parse_loadplan_icm(self, data, kwargs):
        return move.ICMMove(move.ICMMove.LOADPLAN, **kwargs)

    def parse_cardinal_sequencing_icm(self, data, kwargs):
        content_speaker = data.get("content_speaker")
        return move.ICMMoveWithContent(
            move.ICMMove.CARDINAL_SEQUENCING, int(data["content"]), content_speaker=content_speaker, **kwargs
        )

    def parse_report_inference_icm(self, data, kwargs):
        content_data = data.get("content")
        if content_data:
            content = self.parse(content_data)
            content_speaker = data.get("content_speaker")
            return move.ICMMoveWithSemanticContent(
                move.ICMMove.REPORT_INFERENCE, content, content_speaker=content_speaker, **kwargs
            )

    def parse_greet_move(self, kwargs):
        return move.GreetMove(**kwargs)

    def parse_insult_move(self, kwargs):
        return move.InsultMove(**kwargs)

    def parse_insult_response_move(self, kwargs):
        return move.InsultResponseMove(**kwargs)

    def parse_mute_move(self, kwargs):
        return move.MuteMove(**kwargs)

    def parse_unmute_move(self, kwargs):
        return move.UnmuteMove(**kwargs)

    def parse_quit_move(self, kwargs):
        return move.QuitMove(**kwargs)

    def parse_thank_you_move(self, kwargs):
        return move.ThankYouMove(**kwargs)

    def parse_thank_you_response_move(self, kwargs):
        return move.ThankYouResponseMove(**kwargs)

    def parse_prereport_move(self, data):
        service_action = data["service_action"]
        arguments = data["arguments"]
        argument_list = [self.parse_proposition(argument) for argument in arguments]
        ontology_name = data["ontology_name"]
        return move.PrereportMove(ontology_name, service_action, argument_list)

    def parse_report_move(self, data, kwargs):
        content = self.parse(data["content"])
        return move.ReportMove(content, **kwargs)

    def parse_ask_move(self, data, kwargs):
        question = self.parse_question(data["content"])
        return move.AskMove(question, **kwargs)

    def parse_service_action_outcome(self, json):
        if json["service_action_outcome"]:
            return SuccessfulServiceAction()
        return FailedServiceAction(json["failure_reason"])

    def parse_action_status(self, json):
        if json["action_status"] == ActionStatus.DONE:
            return Done()
        if json["action_status"] == ActionStatus.ABORTED:
            return Aborted(json["reason"])


class CheckingJSONParser(NonCheckingJSONParser):
    def __init__(self, ddd_name, ontology, domain_name):
        self.ontology = ontology
        self.ontology_name = ontology.get_name()
        self.domain_name = domain_name

    def parse_action(self, data):
        action_name = data["value"]
        return self.ontology.create_action(action_name)

    def _create_predicate_proposition(self, data):
        if data["predicate"]["sort"]["_name"] == "boolean":
            individual = None
        elif data["predicate"]["sort"]["_name"] == "string" and data["individual"] is None:
            individual = None
        else:
            individual = self.parse_individual(data["individual"])
        predicate = self.parse_predicate(data["predicate"])
        polarity = data.get("_polarity")
        return self.ontology.create_predicate_proposition(predicate, individual, polarity)

    def parse_predicate(self, data):
        predicate_name = data["name"]
        return self.ontology.get_predicate(predicate_name)

    def parse_findout_plan_item(self, data):
        question = self.parse_question(data[plan_item.TYPE_FINDOUT])
        allow_answer_from_pcom = data.get("allow_answer_from_pcom", False)
        return plan_item.Findout(self.domain_name, question, allow_answer_from_pcom)

    def parse_raise_plan_item(self, data):
        question = self.parse_question(data[plan_item.TYPE_RAISE])
        return plan_item.Raise(self.domain_name, question)

    def parse_individual(self, data):
        def extract(string, prefix, suffix):
            if value.startswith(prefix) and value.endswith(suffix):
                return value[len(prefix):len(value) - len(suffix)]

        def extract_datetime_value():
            return extract(value, "datetime(", ")")

        def extract_person_name_value():
            return extract(value, "person_name(", ")")

        if data is None:
            return None
        sort = self.ontology.get_sort(data["sort"]["_name"])
        value = data["value"]
        polarity = data["polarity"]
        if sort.is_datetime_sort():
            value = DateTime(extract_datetime_value())
        if sort.is_real_sort():
            value = float(value)
        if sort.is_person_name_sort():
            value = PersonName(extract_person_name_value())
        if sort.is_integer_sort():
            value = int(value)
        if polarity == Polarity.POS:
            return self.ontology.create_individual(value, sort)
        return self.ontology.create_negative_individual(value)

    def _create_lambda_abstracted_predicate_proposition(self, content):
        predicate = self.parse_predicate(content["predicate"])
        return self.ontology.create_lambda_abstracted_predicate_proposition(predicate)



import argparse
import contextlib
import importlib
import os
import warnings
import logging
import json

from requests.exceptions import MissingSchema
import structlog

from tala.cli import console_formatting
from tala.cli.tdm.tdm_cli import TDMCLI
from tala.config import BackendConfig, DddConfig, DeploymentsConfig, BackendConfigNotFoundException, \
    DddConfigNotFoundException, DeploymentsConfigNotFoundException
from tala.ddd.maker.ddd_maker import DddMaker
from tala.log.logger import configure_stdout_logging, configure_file_logging
from tala.cli.argument_parser import add_common_backend_arguments, add_shared_frontend_and_backend_arguments
from tala.log.formats import VALID_TIS_LOGGING_FORMATS, TIS_LOGGING_AUTO, VALID_TIS_UPDATE_FORMATS, TIS_LOGGING_FULL
from tala.service.constants import DEFAULT_PORT
from tala import domain_orchestration


class UnexpectedDDDException(Exception):
    pass


class ConfigAlreadyExistsException(Exception):
    pass


class ConfigNotFoundException(Exception):
    pass


class InvalidArgumentException(Exception):
    pass


def create_ddd(args):
    DddMaker(args.name, use_rgl=False, target_dir=args.target_dir).make()


def create_backend_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            f"Expected to be able to create backend config file '{args.filename}' but it already exists."
        )
    BackendConfig().write_default_config(args.ddd, args.filename)


def configure_structlog_and_create_logger(args):
    if args.log_to_stdout:
        configure_stdout_logging(args.log_level)
    else:
        configure_file_logging(level=args.log_level)
    return structlog.get_logger("tala")


def create_orchestrated_domain_bundle(args):
    odb_as_json = domain_orchestration.create_odb_as_json(
        args.config, args.overridden_ddd_config_paths, args.should_greet, args.logged_tis_format,
        args.tis_update_format, args.log_to_stdout, args.log_level
    )
    odb = domain_orchestration.OrchestratedDomainBundle(odb_as_json).as_json()
    _store_odb(odb, args.output)


def _store_odb(odb, output_dir):
    print(f"storing orchestrated domain bundle in file '{output_dir}'")
    file = open(output_dir, 'w')
    file.write(json.dumps(odb))
    file.close()


def create_ddd_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            f"Expected to be able to create DDD config file '{args.filename}' but it already exists."
        )
    DddConfig().write_default_config(path=args.filename, use_rgl=False)


def create_deployments_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            f"Expected to be able to create deployments config file '{args.filename}' but it already exists."
        )
    DeploymentsConfig().write_default_config(args.filename)


@contextlib.contextmanager
def _config_exception_handling():
    def generate_message(name, command_name, config):
        return f"Expected {name} config '{config}' to exist but it was not found. To create it, run " \
               f"'tala {command_name} --filename {config}'."

    try:
        yield
    except BackendConfigNotFoundException as exception:
        message = generate_message("backend", "create-backend-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DddConfigNotFoundException as exception:
        message = generate_message("DDD", "create-ddd-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DeploymentsConfigNotFoundException as exception:
        message = generate_message("deployments", "create-deployments-config", exception.config_path)
        raise ConfigNotFoundException(message)


def version(args):
    try:
        tala_version = importlib.metadata.version("tala")
        print(tala_version)
    except importlib.metadata.PackageNotFoundError:
        print("Tala is not installed.")


def interact(args):
    config = DeploymentsConfig(args.deployments_config)
    url = config.get_url(args.environment_or_url)
    resume_session = args.resume_session
    tdm_cli = TDMCLI(url, resume_session)

    try:
        tdm_cli.run()
    except (KeyboardInterrupt, EOFError):
        tdm_cli.stop()
    except MissingSchema:
        environments = list(config.read().keys())
        print(f"Expected a URL or one of the known environments {environments} but got '{url}'")


def add_create_ddd_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd", help="create a new DDD")
    parser.set_defaults(func=create_ddd)
    parser.add_argument("name", help="Name of the DDD, e.g. my_ddd")
    parser.add_argument(
        "--target-dir", default=".", metavar="DIR", help="target directory, will be created if it doesn't exist"
    )


def add_create_odb_subparser(subparsers):
    parser = subparsers.add_parser("create-odb", help="create a new ODB")
    parser.set_defaults(func=create_orchestrated_domain_bundle)

    backend_group = parser.add_argument_group("backend")
    backend_group.add_argument("output", help="Directory where the ODB file will be stored")
    backend_group.add_argument("--should-greet", action="store_true", default=False)
    add_common_backend_arguments(backend_group)
    add_shared_frontend_and_backend_arguments(backend_group)
    logging_group = parser.add_argument_group("logging")
    logging_group.add_argument(
        "--log-to-stdout",
        action="store_true",
        default=False,
        help="If enabled, logging is done to stdout instead of to a file"
    )
    logging_group.add_argument("--port", type=int, default=DEFAULT_PORT, help="listen for connections on this port")
    choices = [TIS_LOGGING_AUTO] + VALID_TIS_LOGGING_FORMATS
    logging_group.add_argument(
        "--tis-format",
        "-tf",
        dest="logged_tis_format",
        choices=choices,
        default=TIS_LOGGING_AUTO,
        help="log TIS on this format"
    )
    logging_group.add_argument(
        "--tis-update-format",
        choices=VALID_TIS_UPDATE_FORMATS,
        default=TIS_LOGGING_FULL,
        help="log TIS updates on this format"
    )


def add_create_backend_config_subparser(subparsers):
    parser = subparsers.add_parser("create-backend-config", help="create a backend config")
    parser.set_defaults(func=create_backend_config)
    parser.add_argument("ddd", help="name of the active DDD, e.g. my_ddd")
    parser.add_argument(
        "--filename",
        default=BackendConfig.default_name(),
        metavar="NAME",
        help=f"filename of the backend config, e.g. {BackendConfig.default_name()}"
    )


def add_create_ddd_config_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd-config", help="create a DDD config")
    parser.set_defaults(func=create_ddd_config)
    parser.add_argument(
        "--filename",
        default=DddConfig.default_name(),
        metavar="NAME",
        help=f"filename of the DDD config, e.g. {DddConfig.default_name()}"
    )


def add_create_deployments_config_subparser(subparsers):
    parser = subparsers.add_parser("create-deployments-config", help="create a deployments config")
    parser.set_defaults(func=create_deployments_config)
    parser.add_argument(
        "--filename",
        default=DeploymentsConfig.default_name(),
        metavar="NAME",
        help=f"filename of the deployments config, e.g. {DeploymentsConfig.default_name()}"
    )


def add_version_subparser(subparsers):
    parser = subparsers.add_parser("version", help="print the Tala version")
    parser.set_defaults(func=version)


def add_deployment_config_arguments(parser):
    parser.add_argument(
        "environment_or_url",
        help="this is either an environment, e.g. 'dev', pointing to a url in the deployments config; "
        "alternatively, if not an environment, this is considered a url in itself; "
        "the url should point to a TDM deployment, e.g. 'https://my-deployment.ddd.tala.cloud:9090/interact'"
    )
    parser.add_argument(
        "--config",
        dest="deployments_config",
        default=None,
        help=f"override the default deployments config {DeploymentsConfig.default_name()!r}"
    )
    parser.add_argument(
        "--resume-session",
        dest="resume_session",
        default=None,
        const="tala-session.state",
        nargs="?",
        help="Session object to be resumed. If parameter omitted, tala will resume latest session."
    )


def add_interact_subparser(subparsers):
    parser = subparsers.add_parser("interact", help="start an interactive chat with a deployed DDD")
    add_deployment_config_arguments(parser)
    parser.set_defaults(func=interact)


def _add_test_arguments(parser):
    add_deployment_config_arguments(parser)
    parser.add_argument(dest="tests_filenames", nargs="+", metavar="TEST-FILE", help="specify DDD test files")
    parser.add_argument("-t", dest="selected_tests", nargs="+", default=[], metavar="TEST", help="select test by name")
    parser.add_argument(
        "-l", "--log-level", nargs="?", default=logging.WARNING, metavar="LOG-LEVEL", help="select logging level"
    )


def format_warnings():
    def warning_on_one_line(message, category, _filename, _lineno, _file=None, _line=None):
        string = f"{category.__name__}: {message}\n"
        return console_formatting.bold(string)

    warnings.formatwarning = warning_on_one_line


def show_deprecation_warnings():
    warnings.simplefilter("always", category=DeprecationWarning)


def main(args=None):
    format_warnings()
    show_deprecation_warnings()
    root_parser = argparse.ArgumentParser(description="Use the Tala SDK for the Talkamatic Dialogue Manager (TDM).")
    subparsers = root_parser.add_subparsers()
    add_create_ddd_subparser(subparsers)
    add_create_odb_subparser(subparsers)
    add_create_backend_config_subparser(subparsers)
    add_create_ddd_config_subparser(subparsers)
    add_create_deployments_config_subparser(subparsers)
    add_version_subparser(subparsers)
    add_interact_subparser(subparsers)

    parsed_args = root_parser.parse_args(args)
    with _config_exception_handling():
        parsed_args.func(parsed_args)


if __name__ == "__main__":
    main()



import unittest

from tala.log.serialization import serialize, unserialize


class CustomClass(object):
    def __eq__(self, other):
        return isinstance(other, CustomClass)

    def __ne__(self, other):
        return not (other == self)


class SerializationTestCase(unittest.TestCase):
    def test_round_trip_for_instance_of_builtin_class(self):
        self.given_serialized("a string")
        self.when_unserializing_serialized_result()
        self.then_result_is("a string")

    def given_serialized(self, obj):
        self._serialized_result = serialize(obj)

    def when_unserializing_serialized_result(self):
        self._result = unserialize(self._serialized_result)

    def then_result_is(self, expected_result):
        self.assertEqual(expected_result, self._result)

    def test_round_trip_for_instance_of_custom_class(self):
        self.given_serialized(CustomClass())
        self.when_unserializing_serialized_result()
        self.then_result_is(CustomClass())

    def test_serialized_string_spans_single_line(self):
        self.when_serialize_string_containing_newline()
        self.then_result_contains_no_newline()

    def when_serialize_string_containing_newline(self):
        self._result = serialize("line1\nline2")

    def then_result_contains_no_newline(self):
        self.assertNotIn(b"\n", self._result)



from unittest.mock import Mock

from tala.config import BackendConfig
from tala.ddd.ddd_manager import DDDManager
from tala.ddd.loading.ddd_loader import DDDLoaderException
from tala.ddd.services.service_interface import ServiceInterface
from tala.testing.ddd_mocker import DddMockingTestCase
from tala.log import logger

from tala.ddd.loading.extended_ddd_loader import ExtendedDDDLoader


class TestExtendedDDDLoader(DddMockingTestCase):
    def setUp(self):
        self.test_logger = logger.configure_and_get_test_logger()
        self._backend_config = BackendConfig.default_config(active_ddd="mockup_app", ddds=["mockup_app"])
        self._mocked_rasa_component_builder = None
        DddMockingTestCase.setUp(self)
        self._mock_warnings = None
        self._MockRasaDDDInterpreter = None
        self._mocked_gf_ddd_interpreter = None
        self._mocked_ontology = None
        self._mocked_grammar = None

    def tearDown(self):
        DddMockingTestCase.tearDown(self)

    def test_load_xml(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._when_load_is_called("mockup_app")
        self._then_result_contains_ddd("mockup_app")

    def _then_result_contains_ddd(self, ddd_name):
        ddd = self._result
        self.assertEqual(ddd_name, ddd.name)
        ddd.ontology
        ddd.domain

    def test_name_field(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._when_load_is_called("mockup_app")
        self._then_loaded_ddd_has_name("mockup_app")

    def _then_loaded_ddd_has_name(self, name):
        self.assertEqual(name, self._result.name)

    def test_domain_field(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._when_load_is_called("mockup_app")
        self._then_loaded_ddd_has_domain("MockupDomain")

    def _then_loaded_ddd_has_domain(self, name):
        self.assertEqual(name, self._result.domain.get_name())

    def test_ontology_field(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._when_load_is_called("mockup_app")
        self._then_loaded_ddd_has_ontology("MockupOntology")

    def _then_loaded_ddd_has_ontology(self, name):
        self.assertEqual(name, self._result.ontology.get_name())

    def test_service_interface_is_loaded(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._when_load_is_called("mockup_app")
        self._then_loaded_ddd_has_service_interface()

    def _then_loaded_ddd_has_service_interface(self):
        self.assertTrue(isinstance(self._result.service_interface, ServiceInterface))

    def test_missing_service_interface_raises_exception(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._when_load_is_called_then_exception_is_raised_matching(
            "mockup_app", DDDLoaderException, "Expected 'service_interface.xml' to exist but it does not"
        )

    def _when_load_is_called_then_exception_is_raised_matching(self, ddd, expected_exception, expected_pattern):
        with self.assertRaisesRegex(expected_exception, expected_pattern):
            self._when_load_is_called(ddd)

    def _when_load_is_called(self, *args, **kwargs):
        self._load(*args, **kwargs)

    def _load(self, ddd_name):
        mock_ddd_manager = Mock(spec=DDDManager)
        mock_ddd_loader = MockExtendedDDDLoader(
            ddd_manager=mock_ddd_manager,
            name=ddd_name,
            ddd_config=self._mock_ddd_config,
            rerank_amount=self._backend_config["rerank_amount"]
        )
        self._result = mock_ddd_loader.load()

    def test_load_without_device_module(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._given_mocked_ddd_config(device_module=None)
        self._when_load_is_called("mockup_app")
        self._then_result_contains_ddd("mockup_app")

    def test_path_field(self):
        self._given_ontology_xml_file("mockup_app/ontology.xml")
        self._given_domain_xml_file("mockup_app/domain.xml")
        self._given_service_interface_xml_file("mockup_app/service_interface.xml")
        self._when_load_is_called("mockup_app")
        self._then_loaded_ddd_has_path("%s/mockup_app" % self._temp_dir)

    def _then_loaded_ddd_has_path(self, path):
        self.assertEqual(path, self._result.path)


class MockExtendedDDDLoader(ExtendedDDDLoader):
    def _load_ddds_as_dict(self):
        return self.ddds_as_dict



import base64
import codecs
import json
import zlib
import binascii
import warnings

# Compression


def compress_string(string_to_compress):
    """
    compress data: create bytes object, zip it and encode as base64.

    """
    bytes_to_compress = bytes(string_to_compress, 'utf-8')
    zipped = zlib.compress(bytes_to_compress, level=9)
    return base64.b64encode(zipped).decode()


def compress_json(element_to_compress):
    """
    compress data: dump as JSON, create bytes object, zip it and encode as base64.

    """
    return compress_string(json.dumps(element_to_compress))


def compress(element):
    warnings.warn(
        "tala.utils.compression.compress() is deprecated. Use tala.utils.compression.compress_json() instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return compress_json(element)


# String decompression


def decompress_string(data):
    """
    decompress data, assuming that it was compressed with the compression function in this module -
    string as bytes, zipped and base64-encoded.
    """
    decoded = codecs.decode(data.encode(), "base64")
    decompressed_bytes = zlib.decompress(decoded)
    return decompressed_bytes.decode('utf-8')


def ensure_decompressed_string(data):
    try:
        return decompress_string(data)
    except (AttributeError, zlib.error, binascii.Error, TypeError):
        return data


# JSON decompression


def decompress_json(data):
    """
    decompress data, assuming that it was compressed with the compression function in this module - dumped
    into a json string, as bytes, zipped and base64-encoded.
    """
    return json.loads(decompress_string(data))


def decompress(data):
    warnings.warn(
        "tala.utils.compression.decompress() is deprecated. Use tala.utils.compression.decompress_json() instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return decompress_json(data)


def ensure_decompressed_json(data):
    try:
        return decompress_json(data)
    except (AttributeError, zlib.error, binascii.Error, TypeError):
        return data


def ensure_decompressed(data):
    warnings.warn(
        "tala.utils.compression.ensure_decompressed() is deprecated. Use tala.utils.compression.ensure_decompressed_json() instead.",
        DeprecationWarning,
        stacklevel=2
    )
    return ensure_decompressed_json(data)



from tala.model.plan_item import PlanItem
from tala.model.semantic_object import SemanticObjectWithContent
from tala.utils.as_semantic_expression import AsSemanticExpressionMixin
from tala.utils.unicodify import unicodify
from tala.utils.json_api import JSONAPIObject


class Hint(SemanticObjectWithContent, AsSemanticExpressionMixin, JSONAPIObject):
    @classmethod
    def create_from_json_api_data(cls, plan_data, included):
        def generate_plan_items():
            for plan_item_key in plan_data["relationships"]["content"]["data"]:
                plan_item_data = included.get_object_from_relationship(plan_item_key)
                item = PlanItem.create_from_json_api_data(plan_item_data, included)
                if item is None:
                    raise BaseException("item is None", item)
                yield item

        return cls(list(generate_plan_items()))

    def __init__(self, plan):
        SemanticObjectWithContent.__init__(self, plan)
        self._plan = plan

    def __eq__(self, other):
        try:
            equality = self.plan == other.plan
            return equality
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self._plan)

    @property
    def plan(self):
        return self._plan

    def __str__(self):
        return "Hint" + unicodify(self.plan)

    def as_json_api_dict(self):
        obj = JSONAPIObject("tala.model.hint.Hint")
        for plan_item in self.plan:
            obj.append_relationship("content", plan_item.as_json_api_dict())
        return obj.as_dict
